from __future__ import print_function

import os
import subprocess
import shutil
import sys

import pandas as pd
import sqlite3

import pytest

pd.options.display.expand_frame_repr = False
pd.options.display.precision = 4
pd.options.display.max_columns = None

DATA_FOLDER = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")

def _run_cmdline(cmdline):
    stdout = cmdline + "\n"
    try:
        stdout += str(subprocess.check_output(cmdline, shell=True,
                                          stderr=subprocess.STDOUT))
    except subprocess.CalledProcessError as error:
        print(error, end="", file=sys.stderr)
        raise
    return stdout

def _run_openswath_decoy_generator(regtest, temp_folder):
    os.chdir(temp_folder)
    data_path = os.path.join(DATA_FOLDER, "library_targets.pqp")
    shutil.copy(data_path, temp_folder)
    cmdline = "easypqp openswath-decoy-generator  --in library_targets.pqp --out library.pqp --method pseudo-reverse"

    stdout = _run_cmdline(cmdline)

    conn = sqlite3.connect("library.pqp")
    protein_table = pd.read_sql_query("SELECT * FROM PROTEIN", conn)
    peptide_table = pd.read_sql_query("SELECT * FROM PEPTIDE", conn)
    precursor_table = pd.read_sql_query("SELECT * FROM PRECURSOR", conn)
    transition_table = pd.read_sql_query("SELECT * FROM TRANSITION", conn)
    conn.close()

    print(protein_table.sort_values("ID"),file=regtest)
    print(peptide_table.sort_values("ID"),file=regtest)
    print(precursor_table.sort_values("ID"),file=regtest)
    print(transition_table.sort_values("ID"),file=regtest)

def test_openswath_decoy_generator(tmpdir, regtest):
    _run_openswath_decoy_generator(regtest, tmpdir.strpath)


import warnings
from .util import timestamped_echo

try:
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

import click
import os
import pathlib
import posixpath, ntpath
import re
import operator
import numpy as np
import pandas as pd

# alignment
from sklearn import preprocessing
import sklearn.isotonic
import sklearn.linear_model
import statsmodels.api as sm
from scipy.interpolate import interp1d

# error rate estimation
try:
  from pyprophet.stats import pemp, qvalue, pi0est
  from pyprophet.ipf import compute_model_fdr
except ModuleNotFoundError:
  pass

# plotting
from scipy.stats import gaussian_kde
from numpy import linspace, concatenate
from seaborn import lmplot

def plot(path, title, targets, decoys):
  plt.figure(figsize=(10, 5))
  plt.subplots_adjust(hspace=.5)

  plt.subplot(121)
  plt.title("group score distributions")
  plt.xlabel("score")
  plt.ylabel("# of groups")
  plt.hist(
      [targets, decoys], 20, color=['g', 'r'], label=['target', 'decoy'], histtype='bar')
  plt.legend(loc=2)

  plt.subplot(122)
  tdensity = gaussian_kde(targets)
  tdensity.covariance_factor = lambda: .25
  tdensity._compute_covariance()
  ddensity = gaussian_kde(decoys)
  ddensity.covariance_factor = lambda: .25
  ddensity._compute_covariance()
  xs = linspace(min(concatenate((targets, decoys))), max(
      concatenate((targets, decoys))), 200)
  plt.title("group score densities")
  plt.xlabel("score")
  plt.ylabel("density")
  plt.plot(xs, tdensity(xs), color='g', label='target')
  plt.plot(xs, ddensity(xs), color='r', label='decoy')
  plt.legend(loc=2)

  plt.suptitle(title)
  plt.savefig(path)
  plt.close()

def peptide_fdr(psms, peptide_fdr_threshold, pi0_lambda, plot_path, nofdr):
  pi0_method = 'bootstrap'
  pi0_smooth_df = 3
  pi0_smooth_log_pi0 = False
  pfdr = False


  if nofdr:
    peptides = psms.groupby(['modified_peptide','decoy','q_value'])['pp'].max().reset_index()
    targets = peptides[~peptides['decoy']].copy()
    decoys = peptides[peptides['decoy']].copy()

  else:
    peptides = psms.groupby(['modified_peptide','decoy'])['pp'].max().reset_index()
    targets = peptides[~peptides['decoy']].copy()
    decoys = peptides[peptides['decoy']].copy()

    targets['p_value'] = pemp(targets['pp'], decoys['pp'])
    targets['q_value'] = qvalue(targets['p_value'], pi0est(targets['p_value'], pi0_lambda, pi0_method, pi0_smooth_df, pi0_smooth_log_pi0)['pi0'], pfdr)

    plot(plot_path, "global peptide scores", targets['pp'], decoys['pp'])
  
  return targets[targets['q_value'] < peptide_fdr_threshold]['modified_peptide'], np.min(targets[targets['q_value'] < peptide_fdr_threshold]['pp'])

def protein_fdr(psms, protein_fdr_threshold, pi0_lambda, plot_path, nofdr):
  pi0_method = 'bootstrap'
  pi0_smooth_df = 3
  pi0_smooth_log_pi0 = False
  pfdr = False

  if nofdr:
    proteins = psms.groupby(['protein_id','decoy','q_value'])['pp'].max().reset_index()
    targets = proteins[~proteins['decoy']].copy()
    decoys = proteins[proteins['decoy']].copy()

  else:
    proteins = psms.groupby(['protein_id','decoy'])['pp'].max().reset_index()  
    targets = proteins[~proteins['decoy']].copy()
    decoys = proteins[proteins['decoy']].copy()

    targets['p_value'] = pemp(targets['pp'], decoys['pp'])
    targets['q_value'] = qvalue(targets['p_value'], pi0est(targets['p_value'], pi0_lambda, pi0_method, pi0_smooth_df, pi0_smooth_log_pi0)['pi0'], pfdr)

    plot(plot_path, "global protein scores", targets['pp'], decoys['pp'])
  
  return targets[targets['q_value'] < protein_fdr_threshold]['protein_id'], np.min(targets[targets['q_value'] < protein_fdr_threshold]['pp'])

def process_psms(psms, psmtsv, peptidetsv, psm_fdr_threshold, peptide_fdr_threshold, protein_fdr_threshold, pi0_lambda, peptide_plot_path, protein_plot_path, proteotypic, nofdr):
  # Append columns
  psms['base_name'] = psms['run_id'].apply(lambda x: os.path.splitext(os.path.basename(x))[0])

  if None not in (psmtsv, peptidetsv):
    # Read psm.tsv and peptide.tsv
    peptidetsv_df = pd.read_csv(peptidetsv, index_col=False, sep='\t', usecols=["Peptide", "Gene", "Protein ID"])
    psmtsv_df = pd.read_csv(psmtsv, index_col=False, sep='\t', usecols=["Spectrum", 'Spectrum File', "Peptide"])

    # Filter out PSMs whose peptides are not in peptide.tsv
    psmtsv_df = psmtsv_df[psmtsv_df["Peptide"].isin(peptidetsv_df["Peptide"])]

    # Generate a group_id column
    temp_df = psmtsv_df["Spectrum"].str.split('.', expand=True)
    psmtsv_df["group_id"] = temp_df.iloc[:, 0] + "_" + pd.to_numeric(temp_df.iloc[:, -2]).astype(str) + \
                            psmtsv_df['Spectrum File'].apply(lambda x: posixpath.basename(ntpath.basename(x))).str.extract('(_rank[0-9]+)', expand=False).apply(lambda x: '' if pd.isna(x) else x)

    # Filter psm dataframe
    psms = psms[psms["group_id"].isin(psmtsv_df["group_id"])]

    # Update gene_id and protein_id
    psms = psms.merge(peptidetsv_df, how="left", left_on="peptide_sequence", right_on="Peptide")
    psms.drop(["gene_id", "protein_id"], inplace=True, axis=1)
    psms.rename(columns={"Gene": "gene_id", "Protein ID": "protein_id"}, inplace=True)
    psms["num_tot_proteins"] = 1
    timestamped_echo("Info: %s redundant PSMs identified after filtering with %s and %s" % (psms.shape[0], psmtsv, peptidetsv))
  else:
    # Filter proteotypic peptides
    if proteotypic:
      psms = psms[psms['num_tot_proteins'] == 1].copy()
    else:
      raise click.ClickException("Support for non-proteotypic peptides is not yet implemented.")
    
    # Generate canonical set of protein identifiers
    proteinset = psms[['peptide_sequence','protein_id']].drop_duplicates()
    proteinset['protein_id'] = proteinset['protein_id'].astype(str)
    proteinset_canonical = proteinset.groupby('peptide_sequence').apply(lambda x: ";".join(sorted(list(set([a for b in x['protein_id'].str.split(';').tolist() for a in b]))))).reset_index(name='protein_id')

    psms = pd.merge(psms.drop(columns='protein_id'), proteinset_canonical, on='peptide_sequence')

    # Prepare PeptideProphet / iProphet results
    if 'q_value' not in psms.columns:
      psms['q_value'] = compute_model_fdr(psms['pep'].values)

    # Confident peptides and protein in global context
    peptides, peptide_pp_threshold = peptide_fdr(psms, peptide_fdr_threshold, pi0_lambda, peptide_plot_path, nofdr)
    timestamped_echo("Info: %s modified peptides identified (q-value < %s; PP threshold = %s)" % (len(peptides), peptide_fdr_threshold, peptide_pp_threshold))
    proteins, protein_pp_threshold = protein_fdr(psms, protein_fdr_threshold, pi0_lambda, protein_plot_path, nofdr)
    timestamped_echo("Info: %s proteins identified (q-value < %s; PP threshold = %s)" % (len(proteins), protein_fdr_threshold, protein_pp_threshold))

    # Filter peptides and proteins
    psms = psms[psms['modified_peptide'].isin(peptides)]
    psms = psms[psms['protein_id'].isin(proteins)]

    # Filter PSMs
    psms = psms[psms['q_value'] < psm_fdr_threshold]

    # Remove decoys
    psms = psms[~psms['decoy']]

    timestamped_echo("Info: %s redundant PSMs identified (q-value < %s; PP threshold = %s)" % (psms.shape[0], psm_fdr_threshold, np.min(1-psms['pep'])))

  return psms

def lowess_iso(x, y, lowess_frac):
  with warnings.catch_warnings():
    warnings.filterwarnings('ignore', message='invalid value encountered in ', category=RuntimeWarning)
    lwf = sm.nonparametric.lowess(y, x.ravel(), frac=lowess_frac)
    while pd.isna(lwf[:, 1]).any():
      lowess_frac *= 2
      lwf = sm.nonparametric.lowess(y, x.ravel(), frac=lowess_frac)
  lwf_x = lwf[:, 0]
  ir = sklearn.isotonic.IsotonicRegression()  # make the regression strictly increasing
  lwf_y = ir.fit_transform(lwf_x, lwf[:, 1])
  mask = np.concatenate([[True], np.diff(lwf_y) != 0])  # remove non increasing points
  try:
    return interp1d(lwf_x[mask], lwf_y[mask], bounds_error=False, fill_value="extrapolate")
  except ValueError as e:
    timestamped_echo(e)
    return interp1d(lwf_x, lwf_y, bounds_error=False, fill_value="extrapolate")

class LowessIsoEstimator:
  def __init__(self, lowess_frac):
    self.lowess_frac = lowess_frac

  def fit(self, x, y):
    self.lwi = lowess_iso(x, y, self.lowess_frac)
    return self

  def get_params(self, deep=False):
    return {'lowess_frac': self.lowess_frac}

  def set_params(self, lowess_frac):
    self.lowess_frac = lowess_frac
    return self

  def score(self, x, y):
    resid = self.lwi(x.ravel()) - y
    return 1 / resid.dot(resid)

  def predict(self, x):
    return self.lwi(x.ravel())

  def __repr__(self):
    return str(self.get_params())

def lowess_iso_predictor(filename, x, y, xpred):
  gsc = sklearn.model_selection.GridSearchCV(LowessIsoEstimator(None), {'lowess_frac': [0.01, 0.02, 0.04, 0.08]},
                                             cv=sklearn.model_selection.KFold(4, shuffle=True, random_state=0),
                                             n_jobs=min(os.cpu_count(), 61))

  gsc.fit(x.reshape(-1, 1), y)
  timestamped_echo(f'Info: {filename}; Lowess fraction used: {gsc.best_params_["lowess_frac"]}.')
  return gsc.best_estimator_.predict(xpred)

def lowess(run, reference_run, xcol, ycol, lowess_frac, psm_fdr_threshold, min_peptides, filename, main_path):
  # Filter alignment data
  run_alignment = run[run['q_value'] < psm_fdr_threshold] \
    if 'q_value' in run else \
    run
  if 'q_value' in reference_run:
    reference_run_alignment = reference_run[reference_run['q_value'] < psm_fdr_threshold]
  else:
    reference_run_alignment = reference_run

  dfm = pd.merge(run_alignment, reference_run_alignment[['modified_peptide','precursor_charge',ycol]], on=['modified_peptide','precursor_charge'])
  timestamped_echo(f'Info: {filename}; Peptide overlap between run and reference: {dfm.shape[0]}.')
  if dfm.shape[0] <= min_peptides:
    timestamped_echo(f'Info: {filename}; Skipping run because not enough peptides could be found for alignment.')
    return pd.DataFrame()

  if dfm.shape[0] < 50:  # use linear regression for small reference size
    linreg = sklearn.linear_model.LinearRegression().fit(dfm[xcol].to_numpy().reshape(-1, 1), dfm[ycol])
    run[ycol] = linreg.predict(run[xcol].to_numpy().reshape(-1, 1))
  else:
    # Fit and apply the lowess model
    run[ycol] = lowess_iso_predictor(filename, dfm[xcol].to_numpy(), dfm[ycol].to_numpy(), run[xcol].to_numpy()) \
      if lowess_frac == 0 else \
      lowess_iso(dfm[xcol].to_numpy(), dfm[ycol].to_numpy(), lowess_frac)(run[xcol].to_numpy())

  # Plot regression
  plt.plot(dfm[xcol].to_numpy(), dfm[ycol].to_numpy(), 'o')
  run1 = run[[xcol, ycol]].sort_values(xcol)
  plt.plot(run1[xcol].to_numpy(), run1[ycol].to_numpy())
  plt.xlabel(xcol)
  plt.ylabel(ycol)
  plt.savefig(os.path.join(main_path, filename + ".pdf"))
  plt.close()
  run1.to_pickle(os.path.join(main_path, filename + ".alignment_pkl"))
  return run


def remove_rank_suffix(x):
  '''

  :param x:
  :return:

  >>> remove_rank_suffix('23aug2017_hela_serum_timecourse_4mz_narrow_6_rank4')
  '23aug2017_hela_serum_timecourse_4mz_narrow_6'
  >>> remove_rank_suffix('23aug2017_hela_serum_timecourse_4mz_narrow_6_rank44')
  '23aug2017_hela_serum_timecourse_4mz_narrow_6'
  >>> remove_rank_suffix('23aug2017_hela_serum_timecourse_4mz_narrow_6')
  '23aug2017_hela_serum_timecourse_4mz_narrow_6'
  '''
  import re
  return re.compile('(.+?)(?:_rank[0-9]+)?').fullmatch(x).group(1)


def unify_modified_peptide_masses(mod_pep, transform=None):
  if not hasattr(mod_pep, 'str'):
    return mod_pep, transform
  if transform is None:
    import collections
    float_list = {ee for e in mod_pep.str.findall('\\[(.+?)\\]') for ee in e}
    d = collections.defaultdict(list)
    current_group = None
    for i, (v, k) in enumerate(sorted((float(e), e) for e in float_list)):
      if current_group is None:
        current_group = d[i]
      else:
        if abs(current_group[-1][0] / v - 1) > 0.001:
          current_group = d[i]
      current_group.append((v, k))
    transform = {s: l[0][1] for _, l in d.items() for f, s in l}

  def transform_func(mo):
    ret = mo.group(0)
    for k, v in transform.items():
      ret = ret.replace(k, v)
    return ret

  return mod_pep.str.replace('(?<=\\[).+?(?=\\])', transform_func, regex=True), transform


def generate(files, outfile, psmtsv, peptidetsv, perform_rt_calibration, rt_referencefile, rt_reference_run_path, rt_filter, perform_im_calibration, im_referencefile, im_reference_run_path, im_filter, psm_fdr_threshold, peptide_fdr_threshold, protein_fdr_threshold, rt_lowess_frac, rt_psm_fdr_threshold, im_lowess_frac, im_psm_fdr_threshold, pi0_lambda, peptide_plot_path, protein_plot_path, min_peptides, proteotypic, consensus, nofdr):
  # Parse input arguments
  psm_files = []
  spectra = []

  if len(files) == 1 and files[0].endswith('.txt'):
    files = pathlib.Path(files[0]).read_text().splitlines()

  for file in files:
    if 'psmpkl' in file:
      psm_files.append(file)
    if 'peakpkl' in file:
      spectra.append(file)

  if len(psm_files) == 0:
    raise click.ClickException("No PSMs files present. Need to have tag 'psmpkl' in filename.")

  if len(spectra) == 0:
    raise click.ClickException("No spectrum files present. Need to have tag 'peakpkl' in filename.")

  if peptidetsv is not None and psmtsv is None:
    raise click.ClickException("There is a peptide.tsv but no psm.tsv.")
  elif peptidetsv is None and psmtsv is not None:
    raise click.ClickException("There is a psm.tsv but no peptide.tsv.")

  if None not in (psmtsv, peptidetsv):
    timestamped_echo("Info: There are psm.tsv and peptide.tsv. Will ignore --psm_fdr_threshold, --peptide_fdr_threshold, --protein_fdr_threshold, --pi0_lambda, --proteotypic, and --no-proteotypic.")

  # Read all PSM files
  psms_list = []
  for psm_file in psm_files:
    timestamped_echo("Info: Reading file %s." % psm_file)
    psm_tab = pd.read_pickle(psm_file)
    if psm_tab.shape[0] > 0:
      psms_list.append(psm_tab)
  psms = pd.concat(psms_list).reset_index(drop=True)
  psms['pp'] = 1-psms['pep']
  psms['modified_peptide'], transform_mass = unify_modified_peptide_masses(psms['modified_peptide'])

  timestamped_echo("Info: In total %s PSMs loaded." % psms.shape[0])

  pepid = process_psms(psms, psmtsv, peptidetsv, psm_fdr_threshold, peptide_fdr_threshold, protein_fdr_threshold, pi0_lambda, peptide_plot_path, protein_plot_path, proteotypic, nofdr)

  # Get main path for figures
  main_path = os.path.dirname(os.path.abspath(peptide_plot_path))

  # Generate set of best replicate identifications per run
  pepidr = pepid.loc[pepid.groupby(['base_name','modified_peptide','precursor_charge'])['pp'].idxmax()].sort_index()

  aligned_runs = pepidr # this variable will store the aligned runs
  # Prepare reference iRT list (if enabled)
  if perform_rt_calibration:
    rt_reference_run_columns = ['modified_peptide', 'precursor_charge', 'irt']

    if rt_referencefile is not None:
      # Read reference file if present
      rt_reference_run = pd.read_csv(rt_referencefile, index_col=False, sep='\t')
      if not set(rt_reference_run_columns).issubset(rt_reference_run.columns):
        raise click.ClickException("Reference iRT file has wrong format. Requires columns 'modified_peptide', 'precursor_charge' and 'irt'.")
      if rt_reference_run.shape[0] < 10:
        raise click.ClickException("Reference iRT file has too few data points. Requires at least 10.")
    else:
      # Select reference run
      pepidr_stats = pepidr.groupby('base_name')[['modified_peptide']].count().reset_index()
      timestamped_echo(pepidr_stats)

      if rt_filter is not None:
        timestamped_echo("Info: Filter candidate RT reference runs by tag '%s'." % rt_filter)
        pepidr_stats = pepidr_stats[pepidr_stats['base_name'].str.contains(rt_filter)]
        timestamped_echo(pepidr_stats)

      rt_reference_run_base_name = pepidr_stats.loc[pepidr_stats['modified_peptide'].idxmax()]['base_name']

      rt_reference_run = pepidr[pepidr['base_name'] == rt_reference_run_base_name].copy()

      # Normalize RT of reference run
      min_max_scaler = preprocessing.MinMaxScaler()
      rt_reference_run['irt'] = min_max_scaler.fit_transform(rt_reference_run[['retention_time']])*100
      rt_reference_run[rt_reference_run_columns].to_csv(rt_reference_run_path, sep='\t', index=False)

    # Normalize RT of all runs against reference
    aligned_runs = aligned_runs.groupby('base_name', as_index=False, group_keys=False).apply(lambda x: lowess(x, rt_reference_run, 'retention_time', 'irt', rt_lowess_frac, rt_psm_fdr_threshold, min_peptides, "easypqp_rt_alignment_" + x.name, main_path))

  else: # in this case no rt_calibration is performed, we just scale the retention time
    aligned_runs = pepidr
    min_max_scaler = preprocessing.MinMaxScaler()
    aligned_runs['irt'] = min_max_scaler.fit_transform(aligned_runs[['retention_time']])*100


  # Determine if IM is present in the search data
  if pepidr['ion_mobility'].isnull().all():
      enable_im = False
  else:
      enable_im = True


  if perform_im_calibration and enable_im:
    # Prepare reference IM list
    im_reference_run_columns = ['modified_peptide', 'precursor_charge', 'im']

    if im_referencefile is not None:
      # Read reference file if present
      im_reference_run = pd.read_csv(im_referencefile, index_col=False, sep='\t')
      if not set(im_reference_run_columns).issubset(im_reference_run.columns):
        raise click.ClickException("Reference IM file has wrong format. Requires columns 'modified_peptide', 'precursor_charge' and 'im'.")
      if im_reference_run.shape[0] < 10:
        raise click.ClickException("Reference IM file has too few data points. Requires at least 10.")

    else:
      # Select reference run
      pepidr_stats = pepidr.groupby('base_name')[['modified_peptide']].count().reset_index()
      timestamped_echo(pepidr_stats)

      if im_filter is not None:
        timestamped_echo("Info: Filter candidate IM reference runs by tag '%s'." % im_filter)
        pepidr_stats = pepidr_stats[pepidr_stats['base_name'].str.contains(im_filter)]
        timestamped_echo(pepidr_stats)

      im_reference_run_base_name = pepidr_stats.loc[pepidr_stats['modified_peptide'].idxmax()]['base_name']

      im_reference_run = pepidr[pepidr['base_name'] == im_reference_run_base_name].copy()

      # Set IM of reference run
      im_reference_run['im'] = im_reference_run['ion_mobility']
      im_reference_run[im_reference_run_columns].to_csv(im_reference_run_path, sep='\t', index=False)

    # perform IM calibration
    aligned_runs = aligned_runs.groupby('base_name', as_index=False).apply(lambda x: lowess(x, im_reference_run, 'ion_mobility', 'im', im_lowess_frac, im_psm_fdr_threshold, min_peptides, "easypqp_im_alignment_" + x.name, main_path))
  
  elif enable_im: # if no calibration just transfer information as is
    aligned_runs['im'] = aligned_runs['ion_mobility']
  else:
    pass

  pepida = aligned_runs

  if pepida.empty or 'irt' not in pepida.columns:
    timestamped_echo('Info: Not enough peptides could be found for alignment. There will be a blank spectral library.')
    return

  # Remove peptides without valid iRT
  pepida = pepida.loc[np.isfinite(pepida['irt'])]

  # Remove peptides without valid IM
  if enable_im:
    pepida = pepida.loc[np.isfinite(pepida['im'])]
  else:
    pepida.loc[:, 'im'] = np.nan

  if pepida.empty:
    timestamped_echo('Info: Not enough peptides could be found for alignment. There will be a blank spectral library.')
    return

  # Generate set of non-redundant global best replicate identifications
  pepidb = pepida.loc[pepida.groupby(['modified_peptide','precursor_charge'])['pp'].idxmax()].sort_index()

  # Prepare ID mzML pairing
  peak_files = pd.DataFrame({'path': spectra})
  peak_files['base_name'] = peak_files['path'].apply(lambda x: remove_rank_suffix(os.path.splitext(os.path.basename(x))[0]))

  # Parse mzXML to retrieve peaks and store results in peak files
  replicate_pqp = []
  for idx, peak_file in peak_files.iterrows():
    timestamped_echo("Info: Parsing file %s." % peak_file['path'])
    meta_run = pepida[pepida['base_name'] == peak_file['base_name']]
    if meta_run.shape[0] > 0:
      meta_global = pepidb[pepidb['base_name'] == peak_file['base_name']]
      peaks = pd.read_pickle(peak_file['path'])
      peaks['modified_peptide'], _ = unify_modified_peptide_masses(peaks['modified_peptide'], transform_mass)
      # Generate run-specific PQP files for OpenSWATH alignment
      if consensus or ("_Q1" in peak_file['base_name']):
        run_pqp = pd.merge(meta_run, peaks, on=['modified_peptide','precursor_charge','scan_id'])[['precursor_mz','product_mz','fragment','intensity','irt','im','protein_id','gene_id','peptide_sequence','modified_peptide','precursor_charge']]
        run_pqp.columns = ['PrecursorMz','ProductMz','Annotation','LibraryIntensity','NormalizedRetentionTime','PrecursorIonMobility','ProteinId','GeneName','PeptideSequence','ModifiedPeptideSequence','PrecursorCharge']
        run_pqp['PrecursorCharge'] = run_pqp['PrecursorCharge'].astype(int)
        run_pqp_path = os.path.splitext(peak_file['path'])[0]+"_run_peaks.tsv"
        run_pqp.to_csv(run_pqp_path, sep="\t", index=False)
        if consensus:
          replicate_pqp.append(run_pqp)

      # Generate global non-redundant PQP files
      if not consensus:
        global_pqp = pd.merge(meta_global, peaks, on=['modified_peptide','precursor_charge','scan_id'])[['precursor_mz','product_mz','fragment','intensity','irt','im','protein_id','gene_id','peptide_sequence','modified_peptide','precursor_charge']]
        global_pqp.columns = ['PrecursorMz','ProductMz','Annotation','LibraryIntensity','NormalizedRetentionTime','PrecursorIonMobility','ProteinId','GeneName','PeptideSequence','ModifiedPeptideSequence','PrecursorCharge']
        global_pqp['PrecursorCharge'] = global_pqp['PrecursorCharge'].astype(int)
        replicate_pqp.append(global_pqp)

  # Aggregate consensus spectra
  pqp = pd.concat(replicate_pqp)
  if consensus:
    pqp_irt = pqp[['ModifiedPeptideSequence','PrecursorCharge','NormalizedRetentionTime','PrecursorIonMobility']].drop_duplicates().groupby(['ModifiedPeptideSequence','PrecursorCharge'])[['NormalizedRetentionTime','PrecursorIonMobility']].median().reset_index()
    pqp_mass = pqp.groupby(['PrecursorMz','ProductMz','Annotation','ProteinId','GeneName','PeptideSequence','ModifiedPeptideSequence','PrecursorCharge'], dropna=False)['LibraryIntensity'].median().reset_index()
    pqp = pd.merge(pqp_mass,pqp_irt, on=['ModifiedPeptideSequence','PrecursorCharge'])

  # Write output TSV file
  pqp.to_csv(outfile, sep="\t", index=False)



import pyopenms as po
import click
from typing import Union, Tuple

from .targetedfileconverter import TargetedExperiment

def check_argument_values(arg_name: str, arg_value: any, expected_type: Tuple[Union[type, None], Union[Tuple, None]]) -> None:
    """
    Check if the given argument value is of the expected type and value range (if applicable).
    Raise a TypeError or ValueError if the value is invalid.
    """
    expected_type, expected_range = expected_type
    if isinstance(expected_type, list) and None in expected_type:
        pass
    elif not isinstance(arg_value, expected_type):
        raise TypeError(f"{arg_name} should be of type {expected_type.__name__} not type {arg_value.__class__}.")
    if expected_range is not None:
        # Handle numeric range
        if isinstance(expected_range, tuple) and len(expected_range) == 2:
            if not (expected_range[0] <= arg_value <= expected_range[1]):
                raise ValueError(f"{arg_name} should be within the range {expected_range}, cannot except {arg_value}.")
        elif isinstance(expected_range, list) and arg_value not in expected_range:
            raise ValueError(f"{arg_name} should be one of {expected_range}, cannot except '{arg_value}'.")

class OpenSwathDecoyGenerator(TargetedExperiment):
    def __init__(self, 
                 infile: str, 
                 outfile: str="library.pqp", 
                 in_type: Union[str, None]=None, 
                 out_type: Union[str, None]=None, 
                 method: str="shuffle",
                 decoy_tag: str="DECOY_",
                 min_decoy_fraction: float=0.8,
                 aim_decoy_fraction: float=1.0,
                 shuffle_max_attempts: int=30,
                 shuffle_sequence_identity_threshold: float=0.5,
                 shift_precursor_mz_shift: float=0.0,
                 shift_product_mz_shift: float=20.0,
                 product_mz_threshold: float=0.025,
                 allowed_fragment_types: str="b,y",
                 allowed_fragment_charges: str="1,2,3,4",
                 enable_detection_specific_losses: bool=False,
                 enable_detection_unspecific_losses: bool=False,
                 switchKR: bool=True,
                 separate: bool=False) -> None:
        super().__init__(True)

        # Valdiate arguments
        check_argument_values("infile", infile, (str, None))
        check_argument_values("outfile", outfile, (str, None))
        # Handle types
        if in_type is None:
            in_type = self._get_file_type(infile)
        if out_type is None:
            out_type = self._get_file_type(outfile)
        check_argument_values("in_type", in_type, ([str, None], ['tsv', 'mrm', 'pqp', 'TraML']))
        check_argument_values("out_type", out_type, ([str, None], ['tsv', 'pqp', 'TraML']))
        check_argument_values("method", method, (str, ['shuffle', 'pseudo-reverse', 'reverse', 'shift']))
        check_argument_values("decoy_tag", decoy_tag, (str, None))
        check_argument_values("min_decoy_fraction", min_decoy_fraction, (float, (0, 1)))
        check_argument_values("aim_decoy_fraction", aim_decoy_fraction, (float, (0, 1)))
        check_argument_values("shuffle_max_attempts", shuffle_max_attempts, (int, None))
        check_argument_values("shuffle_sequence_identity_threshold", shuffle_sequence_identity_threshold, (float, (0, 1)))
        check_argument_values("shift_precursor_mz_shift", shift_precursor_mz_shift, (float, None))
        check_argument_values("shift_product_mz_shift", shift_product_mz_shift, (float, None))
        check_argument_values("product_mz_threshold", product_mz_threshold, (float, None))
        check_argument_values("allowed_fragment_types", allowed_fragment_types, (str, None)) # TODO: Add value check, to ensure valid fragment types
        check_argument_values("allowed_fragment_charges", allowed_fragment_charges, (str, None)) # TODO: Add value check to ensure ints are in string of charges
        check_argument_values("enable_detection_specific_losses", enable_detection_specific_losses, (bool, None))
        check_argument_values("enable_detection_unspecific_losses", enable_detection_unspecific_losses, (bool, None))
        check_argument_values("switchKR", switchKR, (bool, None))
        check_argument_values("separate", separate, (bool, None))

        # TODO: Move this up before argument validation for specific arg?
        # Transform string
        allowed_fragment_types = allowed_fragment_types.split(",")
        allowed_fragment_types = [s.encode('utf-8') for s in allowed_fragment_types]
        allowed_fragment_charges = allowed_fragment_charges.split(",")
        allowed_fragment_charges = [int(charge) for charge in allowed_fragment_charges]

        # Assign values to self
        for name, value in locals().items():
            if name != 'self':
                # print(f"Info: Setting {name} = {value}")
                setattr(self, name, value)

        # Load target experiment
        self.load_library(self.infile, self.in_type)

    def generate_decoys(self) -> None:
        # Initiate decoy experiment
        self.tr_decoy = po.TargetedExperiment()

        # Generate decoys
        decoys = po.MRMDecoy()
        decoys.generateDecoys(self.tr_exp, self.tr_decoy, self.method, self.aim_decoy_fraction, self.switchKR, self.decoy_tag, self.shuffle_max_attempts, self.shuffle_sequence_identity_threshold, self.shift_precursor_mz_shift, self.shift_product_mz_shift, self.product_mz_threshold, self.allowed_fragment_types, self.allowed_fragment_charges, self.enable_detection_specific_losses, self.enable_detection_unspecific_losses, -4)

        click.echo(f"Info: Number of target peptides: {len(self.tr_exp.getPeptides())}")
        click.echo(f"Info: Number of decoy peptides: {len(self.tr_decoy.getPeptides())}")
        click.echo(f"Info: Number of target proteins: {len(self.tr_exp.getProteins())}")
        click.echo(f"Info: Number of decoy proteins: {len(self.tr_decoy.getProteins())}")
        
        if len(self.tr_decoy.getPeptides()) / len(self.tr_exp.getPeptides()) < self.min_decoy_fraction or len(self.tr_decoy.getProteins()) / len(self.tr_exp.getProteins()) < self.min_decoy_fraction:
            raise click.ClickException(f"The number of decoys for peptides or proteins is below the threshold of {(self.min_decoy_fraction * 100)}% of the number of targets.")
        
        if self.separate:
            click.echo(f"Info: Writing only decoys to file: {self.outfile}")
            self.tr_exp = self.tr_decoy
        else:
            click.echo(f"Info: Writing targets and decoys to file: {self.outfile}")
            self.tr_exp += self.tr_decoy

        self.write_library(self.outfile, self.out_type)



import os
import pandas as pd
import pyopenms as po
import ctypes
import click
from typing import Union

class TargetedExperiment:
    """
    Class to load and write a OpenMS TargetedExperiment
    """
    def __init__(self, legacy_traml_id: bool=True) -> None:
        self.legacy_traml_id = legacy_traml_id
        self.tr_exp = po.TargetedExperiment()
        self.file_types = po.FileTypes()

    def _validate_type(self, file: str, file_type: str) -> None:
        """Method to ensure filetype is a known OpenMS compatible transition list file type."""
        if self.file_types.nameToType(file_type) == po.FileType.UNKNOWN and file_type!='parquet':
            click.FileError(filename=file, hint=f"Error: Could not determine file type! {file}")
    
    def _get_file_type(self, infile) -> str:
        """Method to get file type extension from file."""
        return os.path.splitext(infile)[-1].split('.')[-1]

    def _get_file_type_id(self, file_type: str) -> int:
        """Method to get file type id as annotated in OpenMS filetype database."""
        return self.file_types.nameToType(file_type)-1
    
    def load_library(self, infile: str, in_type: Union[str, None]=None) -> None:
        """
        Method to load data from input transition list into an OpenMS TargetedExperiment Object

        Parameters:
            infile: (str) input transition list file to load
            in_type: (str|None) input file type. Default: None. Will be inferred from infile
        """
        if in_type is None:
            in_type = self._get_file_type(infile)
        self._validate_type(infile, in_type)
        # Convert infile str to ctype c char
        c_in_file = ctypes.create_string_buffer(infile.encode())
        if self._get_file_type_id(in_type) == po.FileType.TSV or self._get_file_type_id(in_type) == po.FileType.MRM:
            click.echo("Info: Reading TSV transition list data...")
            tsv_reader = po.TransitionTSVFile()
            tsv_reader.convertTSVToTargetedExperiment(c_in_file.value, self._get_file_type_id(in_type), self.tr_exp)
            tsv_reader.validateTargetedExperiment(self.tr_exp)
        
        elif self._get_file_type_id(self.in_type) == po.FileType.PQP:
            click.echo("Info: Reading PQP transition list data...")
            pqp_reader = po.TransitionPQPFile()
            pqp_reader.convertPQPToTargetedExperiment(c_in_file.value, self.tr_exp, self.legacy_traml_id)
            pqp_reader.validateTargetedExperiment(self.tr_exp)

        elif self._get_file_type_id(in_type) == po.FileType.TRAML:
            click.echo("Info: Reading TraML transition list data...")
            traml_reader = po.TraMLFile()
            traml_reader.load(c_in_file.value, self.tr_exp)
        
        click.echo(f"Info: Loaded {len(self.tr_exp.getCompounds())} Compounds, {len(self.tr_exp.getProteins()) } Proteins, {len(self.tr_exp.getPeptides())} Peptides, and {len(self.tr_exp.getTransitions())} Transitions")
    
    def write_library(self, outfile: str, out_type: Union[str, None]=None) -> None:
        """
        Method to write data from an OpenMS TargetedExperiment Object to disk

        Parameters:
            outfile: (str) output transition list file to load
            out_type: (str|None) output file type Default: None. Will be inferred from infile
        """
        if out_type is None:
            out_type = self._get_file_type(outfile)
        self._validate_type(outfile, out_type)
        # Convert outfile str to ctype c char
        c_out_file = ctypes.create_string_buffer(outfile.encode())
        if self._get_file_type_id(out_type) == po.FileType.TSV:
            click.echo("Info: Writing TSV transition list data to disk...")
            tsv_reader = po.TransitionTSVFile()
            self.tr_exp.getPeptides()
            tsv_reader.convertTargetedExperimentToTSV(c_out_file.value, self.tr_exp)

        elif self._get_file_type_id(out_type) == po.FileType.PQP:
            click.echo("Info: Writing PQP transition list data to disk...")
            pqp_reader = po.TransitionPQPFile()
            pqp_reader.convertTargetedExperimentToPQP(c_out_file.value, self.tr_exp)

        elif self._get_file_type_id(out_type) == po.FileType.TRAML:
            click.echo("Info: Writing TraML transition list data to disk...")
            traml_reader = po.TraMLFile()
            traml_reader.store(c_out_file.value, self.tr_exp)



class TargetedFileConverter(TargetedExperiment):
    '''
    TargetedFileConverter

    Converts different spectral libraries / transition files for targeted proteomics and metabolomics analysis.
  
    Can convert multiple formats to and from TraML (standardized transition format). The following formats are supported:
    
        - @ref OpenMS::TraMLFile "TraML" 
        - @ref OpenMS::TransitionTSVFile "OpenSWATH TSV transition lists" 
        - @ref OpenMS::TransitionPQPFile "OpenSWATH PQP SQLite files" 
        - SpectraST MRM transition lists 
        - Skyline transition lists 
        - Spectronaut transition lists 
        - Parquet transition lists 
    '''

    def __init__(self, infile: str, outfile: str="library.pqp", in_type: Union[str, None]=None, out_type: Union[str, None]=None, legacy_traml_id: bool=True) -> None:
        super().__init__(legacy_traml_id)
        self.infile = infile
        self.outfile = outfile

        # Handle types
        if in_type is None:
            in_type = self._get_file_type(self.infile)
        self.in_type = in_type
        if out_type is None:
            out_type = self._get_file_type(self.outfile)
        self.out_type = out_type

    def convert(self) -> None:
        """Method for converting between spectral library formats"""
        # If input is parquet, need to write out a temporary tsv to consume for conversion
        if self.in_type == 'parquet':
            tr_list = pd.read_parquet(self.infile)
            # Write out a temp tsv file for loading into a TargetedExperiment Object
            temp_in_tsv = f"{os.path.splitext(self.infile)[0]}.tsv"
            tr_list.to_csv(temp_in_tsv, sep="\t")
            # Save org infile information
            self.infile_parquet = self.infile
            self.in_type_parquet = self.in_type
            # Overwrite org infile information with TSV information
            self.infile = temp_in_tsv
            self.in_type = "tsv"    

        # Read Input into TargetedExperiment
        self.load_library(self.infile, self.in_type)

        # Write TargetedExperiment to Output
        self.write_library(self.outfile, self.out_type)

        # Clean Up
        if 'in_type_parquet' in dir(self) and self.out_type!='tsv':
            os.remove(self.infile)
            self.infile = self.infile_parquet

        click.echo(f"Info: Finished converting {self.infile} to {self.outfile}")


import os
import pandas as pd
import pyopenms as po
import ctypes
import click
from typing import Union, Tuple

from .targetedfileconverter import TargetedExperiment

def check_argument_values(arg_name: str, arg_value: any, expected_type: Tuple[Union[type, None], Union[Tuple, None]]) -> None:
    """
    Check if the given argument value is of the expected type and value range (if applicable).
    Raise a TypeError or ValueError if the value is invalid.
    """
    expected_type, expected_range = expected_type
    if isinstance(expected_type, list) and None in expected_type:
        pass
    elif not isinstance(arg_value, expected_type):
        raise TypeError(f"{arg_name} should be of type {expected_type.__name__} not type {arg_value.__class__}.")
    if expected_range is not None:
        # Handle numeric range
        if isinstance(expected_range, tuple) and len(expected_range) == 2:
            if not (expected_range[0] <= arg_value <= expected_range[1]):
                raise ValueError(f"{arg_name} should be within the range {expected_range}, cannot except {arg_value}.")
        elif isinstance(expected_range, list) and arg_value not in expected_range:
            raise ValueError(f"{arg_name} should be one of {expected_range}, cannot except '{arg_value}'.")

def check_fragment_type(input_str: str):
    possible_fragment_types = ['b','y','a','x','c','z']
    if input_str not in possible_fragment_types:
        raise ValueError(f"{input_str} is not one of the possible fragment types {possible_fragment_types}")

def string_to_list(input_str: str, output_type: type):
    str_list = input_str.split(",")
    ret_list = []
    for s in str_list:
        if (output_type == bytes):
            check_fragment_type(s)
            convert = bytes(s,encoding='utf-8' )
        else:
            convert = int(s)
        ret_list.append(convert) 

    return ret_list

def read_swath_file(file: str):
    click.echo("Validate provided Swath windows file:")
    swath_window_loader = po.SwathWindowLoader()
    swath_prec_lower = []
    swath_prec_upper = []
    ret_val = []
    swath_window_loader.readSwathWindows(file, swath_prec_lower, swath_prec_upper)
    click.echo("Read Swath maps file with %s windows" % str(len(swath_prec_lower)))
    for idx, s in enumerate(swath_prec_lower):
        current_win = []
        current_win.append(s)
        current_win.append(swath_prec_upper[idx])
        click.echo("Read lower swath window %s and upper window %s" % (s, swath_prec_upper[idx]))
        ret_val.append(current_win)
    return ret_val

def read_unimod_file(unimod_file):
    ### TODO
    return None
    # mods_database = po.ModificationsDB(unimod_file)
    
    # click.echo("Unimod XML: %s modification types and residue specificities imported from file: %s" % (mods_database.getNumberOfModifications(), unimod_file))

class OpenSwathAssayGenerator(TargetedExperiment):
    def __init__(self, infile, in_type, outfile, out_type, min_transitions, max_transitions, allowed_fragment_type, allowed_fragment_charges, enable_detection_specific_losses, enable_detection_unspecific_losses, precursor_mz_threshold, precursor_lower_mz_limit,
                 precursor_upper_mz_limit, product_mz_threshold, product_lower_mz_limit, product_upper_mz_limit, swath_windows_file, unimod_file, enable_ipf, max_num_alternative_localizations, disable_identification_ms2_precursors, disable_identification_specific_losses, enable_identification_unspecific_losses, enable_swath_specifity) -> None:
        super().__init__(True)

        self.infile = infile 
        self.in_type = in_type

        self.outfile = outfile
        self.out_type = out_type 

        self.min_transitions = min_transitions
        self.max_transitions = max_transitions

        self.allowed_fragment_type = string_to_list(allowed_fragment_type, bytes) 

        self.allowed_fragment_charges = string_to_list(allowed_fragment_charges, int)  ### TODO: check valid fragment charges

        self.enable_detection_specific_losses = enable_detection_specific_losses
        self.enable_detection_unspecific_losses = enable_detection_unspecific_losses
        self.precursor_mz_threshold = precursor_mz_threshold
        self.precursor_lower_mz_limit = precursor_lower_mz_limit
        self.precursor_upper_mz_limit = precursor_upper_mz_limit
        self.product_mz_threshold = product_mz_threshold
        self.product_lower_mz_limit = product_lower_mz_limit
        self.product_upper_mz_limit = product_upper_mz_limit

        self.swathes = list(list()) if swath_windows_file == None else read_swath_file(swath_windows_file)
       
       
        ### TODO: read unimod file 
        self.unimod_file = None if unimod_file == None else read_unimod_file(unimod_file)
        print(self.unimod_file )
        ### TODO: implement enable ipf
        self.enable_ipf = enable_ipf
        self.max_num_alternative_localizations = max_num_alternative_localizations
        self.disable_identification_ms2_precursors = disable_identification_ms2_precursors
        self.disable_identification_specific_losses = disable_identification_specific_losses
        self.enable_identification_unspecific_losses = enable_identification_unspecific_losses
        self.enable_swath_specifity = enable_swath_specifity



        ### check argument
        # # Valdiate arguments
        # check_argument_values("infile", infile, (str, None))
        # check_argument_values("outfile", outfile, (str, None))
        # # Handle types
        # if in_type is None:
        #     in_type = self._get_file_type(infile)
        # if out_type is None:
        #     out_type = self._get_file_type(outfile)
        # check_argument_values("in_type", in_type, ([str, None], ['tsv', 'mrm', 'pqp', 'TraML']))
        # check_argument_values("out_type", out_type, ([str, None], ['tsv', 'pqp', 'TraML']))
        # check_argument_values("product_mz_threshold", product_mz_threshold, (float, None))
        # check_argument_values("allowed_fragment_types", allowed_fragment_types, (str, None)) # TODO: Add value check, to ensure valid fragment types
        # check_argument_values("allowed_fragment_charges", allowed_fragment_charges, (str, None)) # TODO: Add value check to ensure ints are in string of charges
        # check_argument_values("enable_detection_specific_losses", enable_detection_specific_losses, (bool, None))
        # check_argument_values("enable_detection_unspecific_losses", enable_detection_unspecific_losses, (bool, None))

    def read_input_file(self) -> None:
        self.load_library(self.infile, self.in_type)
        ### convert to tsv (panda df)

        ### get all transtion for specific precursors



    def annotate_transitions(self) -> None:
        click.echo("Info: Annotating transitions")
        assays = po.MRMAssay()
        assays.reannotateTransitions(self.tr_exp, self.precursor_mz_threshold, self.product_mz_threshold, self.allowed_fragment_type, self.allowed_fragment_charges, self.enable_detection_specific_losses, self.enable_detection_unspecific_losses, -4) ### todo convert fragment type to bytes

        click.echo("Info: Annotating detecting transitions")
        assays.restrictTransitions(self.tr_exp, self.product_lower_mz_limit, self.product_upper_mz_limit, self.swathes) 
        assays.detectingTransitions(self.tr_exp, self.min_transitions, self.max_transitions)

    def write_output_file(self) -> None:
        self.write_library(self.outfile, self.out_type)


import itertools
import pathlib
import pdb
import time
from .util import timestamped_echo
import numpy as np
import pandas as pd
import os
import posixpath, ntpath
from statistics import median_low
import click
import re

# Unimod parsing
import xml.etree.cElementTree as ET
from xml.etree.cElementTree import iterparse

# mzXML parsing
import pyopenms as po


class psmtsv:
	relevant_psm_columns = ["Spectrum",
							"Spectrum File",
							"Peptide",
							"Charge",
							"Retention",
							"Delta Mass",
							"Assigned Modifications",
							"Hyperscore",
							"Nextscore",
							"Expectation",
							"Probability",
							"Ion Mobility",
							"Protein",
							"Protein ID",
							"Gene",
							"Mapped Proteins",
							"Mapped Genes",
							"Glycan q-value",
							]
	rank_pattern = re.compile(r"_rank([\d]+)\.pep\.xml")

	def __init__(self, psmtsv_file, unimod, base_name, exclude_range, enable_unannotated, ignore_unannotated, enable_massdiff, decoy_prefix, labile_mods, max_glycan_q):
		self.psmtsv_file = psmtsv_file
		self.base_name = base_name
		self.decoy_prefix = decoy_prefix
		self.labile_mods = labile_mods
		self.max_glycan_q = max_glycan_q
		self.psms = self.parse_psmtsv()
		self.exclude_range = exclude_range
		self.enable_unannotated = enable_unannotated
		self.ignore_unannotated = ignore_unannotated
		self.enable_massdiff = enable_massdiff
		self.match_unimod(unimod)

	def get(self):
		return(self.psms)

	def parse_psmtsv(self):
		# read relevant PSM table columns
		psms = pd.read_csv(self.psmtsv_file, index_col=False, sep='\t', usecols=lambda col: col in set(psmtsv.relevant_psm_columns))

		psms = psms.apply(self.parse_psm_info, axis=1)
		if self.max_glycan_q != 1:
			if "Glycan q-value" in psms.columns:
				psms = psms[(psms['Glycan q-value'] < self.max_glycan_q) | (psms['Glycan q-value'].isnull())]		# include null to include non-glyco peptides
			else:
				timestamped_echo("Warning: glycan q-value filtering requested, but PSM table does not contain the Glycan q-value column. No filtering performed")
		psms = psms.rename(columns={'Charge': 'precursor_charge',
									'Retention': 'retention_time',
									'Delta Mass': 'massdiff',
									'Peptide': 'peptide_sequence',
									'Ion Mobility': 'ion_mobility',
									'Hyperscore': 'var_hyperscore',
									'Nextscore': 'var_nextscore',
									'Expectation': 'var_expect',
									'Probability': 'pep'
									})
		psms['pep'] = 1 - psms['pep']
		if 'ion_mobility' not in psms:
			psms['ion_mobility'] = np.nan
		if 'hit_rank' not in psms:
			psms['hit_rank'] = 1
		psms = psms.drop(columns=['Spectrum', 'Spectrum File', 'Assigned Modifications', 'Protein', 'Gene', 'Mapped Proteins', 'Mapped Genes', 'Protein ID'])
		if 'Glycan q-value' in psms.columns:
			psms.drop(columns=['Glycan q-value'])
		return psms

	def match_unimod(self, unimod):
		"""
		Match modifications to Unimod as in pepxml. Supports labile modifications
		"""
		def match_modifications(peptide, um, modified_peptide_attribute):
			modified_peptide = peptide['peptide_sequence']

			# parse terminal modifications
			nterm_modification = ""
			if peptide['nterm_modification'] != '':
				nterm_modification = peptide['nterm_modification']
			cterm_modification = ""
			if peptide['cterm_modification'] != '':
				cterm_modification = peptide['cterm_modification']

			# parse closed modifications
			modifications = {}
			if "M|" in peptide[modified_peptide_attribute]:
				for modification in peptide[modified_peptide_attribute].split('|')[1:]:
					site, mass = modification.split('$')
					delta_mass = float(mass)		# psm.tsv mod masses are just the mod, does not include residue mass
					modifications[int(site)] = delta_mass

			massdiff = float(peptide['massdiff'])
			if self.enable_massdiff and (massdiff < self.exclude_range[0] or massdiff > self.exclude_range[1]):
				# parse open modifications
				oms_sequence = peptide['peptide_sequence']
				for site in modifications.keys():
					oms_sequence = oms_sequence[:site-1] + "_" + oms_sequence[site:]

				oms_modifications, nterm_modification, cterm_modification = um.get_oms_id(oms_sequence, peptide['massdiff'], nterm_modification, cterm_modification)
				modifications = {**modifications, **oms_modifications}

			peptide_sequence = peptide['peptide_sequence']
			peptide_length = len(peptide_sequence)
			for site in sorted(modifications, reverse=True):
				positions = ('Anywhere', 'Any N-term', 'Protein N-term') if site == 1 else \
					('Anywhere', 'Any C-term', 'Protein C-term') if site == peptide_length else \
						'Anywhere'
				record_id0 = um.get_id(peptide_sequence[site - 1], positions, modifications[site])
				if isinstance(record_id0, tuple):
					record_id, position = record_id0
				else:
					record_id = record_id0
				is_N_term = isinstance(record_id0, tuple) and position in ('Any N-term', 'Protein N-term')
				if record_id == -1:
					if self.ignore_unannotated:
						return ''		# set empty modified peptide to ignore with later filtering
					elif self.enable_unannotated:
						modified_peptide = "[" + str(round(modifications[site], 6)) + "]" + modified_peptide \
						if is_N_term else \
						modified_peptide[:site] + "[" + str(round(modifications[site], 6)) + "]" + modified_peptide[site:]
					else:
						raise click.ClickException("Error: Could not annotate site %s (%s) from peptide %s with delta mass %s." % (site, peptide['peptide_sequence'][site-1], peptide['peptide_sequence'], modifications[site]))
				else:
					modified_peptide = "(UniMod:" + str(record_id) + ")" + modified_peptide \
						if is_N_term else \
						modified_peptide[:site] + "(UniMod:" + str(record_id) + ")" + modified_peptide[site:]

			if nterm_modification != '':
				record_id_nterm = um.get_id("N-term", 'Any N-term', nterm_modification)
				if record_id_nterm == -1:
					record_id_nterm = um.get_id("N-term", 'Protein N-term', nterm_modification)

				if record_id_nterm == -1:
					if self.ignore_unannotated:
						return ''		# set empty modified peptide to ignore with later filtering
					elif self.enable_unannotated:
						modified_peptide = f'.[{round(nterm_modification, 6)}]{modified_peptide}'
					else:
						raise click.ClickException("Error: Could not annotate N-terminus from peptide %s with delta mass %s." % (peptide['peptide_sequence'], nterm_modification))
				else:
					modified_peptide = ".(UniMod:" + str(record_id_nterm) + ")" + modified_peptide

			if cterm_modification != '':
				record_id_cterm = um.get_id("C-term", 'Any C-term', cterm_modification)
				if record_id_cterm == -1:
					record_id_cterm = um.get_id("C-term", 'Protein C-term', cterm_modification)

				if record_id_cterm == -1:
					if self.ignore_unannotated:
						return ''		# set empty modified peptide to ignore with later filtering
					elif self.enable_unannotated:
						modified_peptide = f'{modified_peptide}.[{round(cterm_modification, 6)}]'
					else:
						raise click.ClickException("Error: Could not annotate C-terminus from peptide %s with delta mass %s." % (peptide['peptide_sequence'], cterm_modification))
				else:
					modified_peptide = modified_peptide + ".(UniMod:" + str(record_id_cterm) + ")"

			return modified_peptide

		if self.psms.shape[0] > 0:
			self.psms['modified_peptide'] = self.psms[['peptide_sequence','modifications','nterm_modification','cterm_modification','massdiff']].apply(lambda x: match_modifications(x, unimod, 'modifications'), axis=1)
			self.psms['labile_modified_peptide'] = self.psms[['peptide_sequence','nonlabile_modifications','nterm_modification','cterm_modification','massdiff']].apply(lambda x: match_modifications(x, unimod, 'nonlabile_modifications'), axis=1)
			if self.ignore_unannotated:
				pre_size = len(self.psms)
				self.psms = self.psms[self.psms['modified_peptide'] != '']
				timestamped_echo("Info: Ignored %s PSMs with modifications not matched to Unimod" % (pre_size - len(self.psms)))

	def parse_psm_info(self, psm_series):
		"""
		Perform parsing operations on a PSM entry
		"""
		psm_series = self.parse_spectrum(psm_series)
		psm_series = self.parse_rank(psm_series)
		psm_series = self.parse_assigned_modifications(psm_series)
		psm_series = self.parse_protein_and_gene(psm_series, self.decoy_prefix)
		return psm_series

	def parse_spectrum(self, psm_series):
		splits = psm_series['Spectrum'].split('.')
		psm_series['run_id'] = splits[0]
		psm_series['scan_id'] = int(splits[1])
		return psm_series

	def parse_rank(self, psm_series):
		rank_match = re.search(psmtsv.rank_pattern, psm_series['Spectrum File'])
		if rank_match:
			psm_series['hit_rank'] = int(rank_match.group(1))
		return psm_series

	def parse_assigned_modifications(self, psm_series):
		"""
		Parse the Assigned Modifications column of a psm.tsv table to easyPQP mods.
		Usage: psm_df = psm_df.apply(method, axis=1), where psm_df is the input PSMs dataframe
		Adds the required modification columns to the dataframe
		"""
		modifications = "M"
		nonlabile_modifications = "M"
		nterm_modification = ""
		cterm_modification = ""
		if pd.notnull(psm_series['Assigned Modifications']):
			mods = psm_series['Assigned Modifications'].split(',')
			for mod in mods:
				match = re.search(r"\((-?\d+\.\d+)\)", mod)
				if match:
					mass = match.group(1)
				else:
					click.echo(
						"Error: invalid modification {} in spectrum {} was ignored".format(mod, psm_series['Spectrum']))
					continue

				if 'N-term' in mod:
					nterm_modification = float(mass)
				elif 'C-term' in mod:
					cterm_modification = float(mass)
				else:
					# regular mod
					splits = mod.split('(')
					location = re.search(r"(\d+)", splits[0]).group(1)
					modifications += '|{}${}'.format(location, mass)
					if self.labile_mods != '':
						if self.labile_mods == 'oglyc':
							if not(float(mass) > 140 and psm_series['Peptide'][int(location) - 1] in ['S', 'T']):
								nonlabile_modifications += '|{}${}'.format(location, mass)
						elif self.labile_mods == 'nglyc':
							if not(float(mass) > 140 and psm_series['Peptide'][int(location) - 1] in ['N']):
								nonlabile_modifications += '|{}${}'.format(location, mass)
						elif self.labile_mods == 'nglyc+':
							# hard code HexNAc remainder mass as the modification mass for fragment ions rather than full modification mass
							if not(float(mass) > 140 and psm_series['Peptide'][int(location) - 1] in ['N']):
								nonlabile_modifications += '|{}${}'.format(location, 203.07937)
					else:
						nonlabile_modifications += '|{}${}'.format(location, mass)
		psm_series['modifications'] = modifications
		psm_series['nonlabile_modifications'] = nonlabile_modifications
		psm_series['nterm_modification'] = nterm_modification
		psm_series['cterm_modification'] = cterm_modification
		return psm_series

	def parse_protein_and_gene(self, psm_series, decoy_prefix):
		"""
		Parse protein and gene IDs from identified and mapped entries and set total_num_proteins and decoy status.
		"""
		psm_series['decoy'] = psm_series['Protein'].startswith(decoy_prefix)

		protein_id = psm_series['Protein ID']
		if pd.notnull(psm_series['Gene']):
			gene_id = psm_series['Gene']
		else:
			gene_id = ''		# contaminants may have empty Gene entry. Ensure string format
		num_total_proteins = 1
		uniprot_like_ids = '|' in psm_series['Protein']

		# if mapped proteins/genes not null, capture all entries
		if pd.notnull(psm_series['Mapped Proteins']):
			splits = psm_series['Mapped Proteins'].split(',')
			for split in splits:
				num_total_proteins += 1
				# If UniProt-like IDs, take ID portion only. Otherwise, use whole ID
				if uniprot_like_ids:
					uniprot_splits = split.split('|')
					protein_id += ';{}'.format(uniprot_splits[1])
				else:
					protein_id += ';{}'.format(split)
		if pd.notnull(psm_series['Mapped Genes']):
			splits = psm_series['Mapped Genes'].split(',')
			for split in splits:
				gene_id += ';{}'.format(split)

		psm_series['protein_id'] = protein_id
		psm_series['gene_id'] = gene_id
		return psm_series


class pepxml:
	def __init__(self, pepxml_file, unimod, base_name, exclude_range, enable_unannotated, enable_massdiff):
		self.pepxml_file = pepxml_file
		self.base_name = base_name
		self.psms = self.parse_pepxml()
		self.exclude_range = exclude_range
		self.enable_unannotated = enable_unannotated
		self.enable_massdiff = enable_massdiff
		self.match_unimod(unimod)

	def get(self):
		return(self.psms)

	def match_unimod(self, unimod):
		def match_modifications(um, peptide):
			monomeric_masses = {"A": 71.03711, "R": 156.10111, "N": 114.04293, "D": 115.02694, "C": 103.00919, "E": 129.04259, "Q": 128.05858, "G": 57.02146, "H": 137.05891, "I": 113.08406, "L": 113.08406, "K": 128.09496, "M": 131.04049, "F": 147.06841, "P": 97.05276, "S": 87.03203, "T": 101.04768, "W": 186.07931, "Y": 163.06333, "V": 99.06841,
								'U': 150.95363, 'O': 237.14773, 'B': 0, 'J': 0, 'X': 0, 'Z': 0}
			modified_peptide = peptide['peptide_sequence']

			# parse terminal modifications
			nterm_modification = ""
			if peptide['nterm_modification'] != '':
				nterm_modification = peptide['nterm_modification'] - 1.0078
			cterm_modification = ""
			if peptide['cterm_modification'] != '':
				cterm_modification = peptide['cterm_modification'] - 18.0153

			# parse closed modifications
			modifications = {}
			if "M|" in peptide['modifications']:
				for modification in peptide['modifications'].split('|')[1:]:
					site, mass = modification.split('$')
					delta_mass = float(mass) - monomeric_masses[peptide['peptide_sequence'][int(site)-1]]
					modifications[int(site)] = delta_mass

			massdiff = float(peptide['massdiff'])
			if self.enable_massdiff and (massdiff < self.exclude_range[0] or massdiff > self.exclude_range[1]):
				# parse open modifications
				oms_sequence = peptide['peptide_sequence']
				for site in modifications.keys():
					oms_sequence = oms_sequence[:site-1] + "_" + oms_sequence[site:]

				oms_modifications, nterm_modification, cterm_modification = um.get_oms_id(oms_sequence, peptide['massdiff'], nterm_modification, cterm_modification)
				modifications = {**modifications, **oms_modifications}

			peptide_sequence = peptide['peptide_sequence']
			peptide_length = len(peptide_sequence)
			for site in sorted(modifications, reverse=True):
				positions = ('Anywhere', 'Any N-term', 'Protein N-term') if site == 1 else \
					('Anywhere', 'Any C-term', 'Protein C-term') if site == peptide_length else \
						'Anywhere'
				record_id0 = um.get_id(peptide_sequence[site - 1], positions, modifications[site])
				if isinstance(record_id0, tuple):
					record_id, position = record_id0
				else:
					record_id = record_id0
				is_N_term = isinstance(record_id0, tuple) and position in ('Any N-term', 'Protein N-term')
				if record_id == -1:
					if self.enable_unannotated:
						modified_peptide = "[" + str(round(modifications[site], 6)) + "]" + modified_peptide \
						if is_N_term else \
						modified_peptide[:site] + "[" + str(round(modifications[site] + monomeric_masses[modified_peptide[site - 1]], 6)) + "]" + modified_peptide[site:]
					else:
						raise click.ClickException("Error: Could not annotate site %s (%s) from peptide %s with delta mass %s." % (site, peptide['peptide_sequence'][site-1], peptide['peptide_sequence'], modifications[site]))
				else:
					modified_peptide = "(UniMod:" + str(record_id) + ")" + modified_peptide \
						if is_N_term else \
						modified_peptide[:site] + "(UniMod:" + str(record_id) + ")" + modified_peptide[site:]

			if nterm_modification != '':
				record_id_nterm = um.get_id("N-term", 'Any N-term', nterm_modification)
				if record_id_nterm == -1:
					record_id_nterm = um.get_id("N-term", 'Protein N-term', nterm_modification)

				if record_id_nterm == -1:
					if self.enable_unannotated:
						modified_peptide = f'.[{round(nterm_modification, 6)}]{modified_peptide}'
					else:
						raise click.ClickException("Error: Could not annotate N-terminus from peptide %s with delta mass %s." % (peptide['peptide_sequence'], nterm_modification))
				else:
					modified_peptide = ".(UniMod:" + str(record_id_nterm) + ")" + modified_peptide

			if cterm_modification != '':
				record_id_cterm = um.get_id("C-term", 'Any C-term', cterm_modification)
				if record_id_cterm == -1:
					record_id_cterm = um.get_id("C-term", 'Protein C-term', cterm_modification)

				if record_id_cterm == -1:
					if self.enable_unannotated:
						modified_peptide = f'{modified_peptide}.[{round(cterm_modification, 6)}]'
					else:
						raise click.ClickException("Error: Could not annotate C-terminus from peptide %s with delta mass %s." % (peptide['peptide_sequence'], cterm_modification))
				else:
					modified_peptide = modified_peptide + ".(UniMod:" + str(record_id_cterm) + ")"


			return modified_peptide

		if self.psms.shape[0] > 0:
			self.psms['modified_peptide'] = self.psms[['peptide_sequence','modifications','nterm_modification','cterm_modification','massdiff']].apply(lambda x: match_modifications(unimod, x), axis=1)

	def parse_pepxml(self):
		peptides = []
		namespaces = {'pepxml_ns': "http://regis-web.systemsbiology.net/pepXML"}
		ET.register_namespace('', "http://regis-web.systemsbiology.net/pepXML")

		context = iterparse(self.pepxml_file, events=("end",))

		for event, elem in context:
			if elem.tag == "{http://regis-web.systemsbiology.net/pepXML}msms_run_summary":
				base_name = basename_spectralfile(posixpath.basename(ntpath.basename(elem.attrib['base_name'])))

				# only proceed if base_name matches
				if base_name == self.base_name:
					# find decoy prefix
					decoy_prefix = ""
					for search_summary in elem.findall('.//pepxml_ns:search_summary', namespaces):
						for parameter in search_summary.findall('.//pepxml_ns:parameter', namespaces):
							if parameter.attrib['name'] == 'decoy_prefix':
								decoy_prefix = parameter.attrib['value']

					# go through all spectrum queries
					for spectrum_query in elem.findall('.//pepxml_ns:spectrum_query', namespaces):
						index = spectrum_query.attrib['index']
						start_scan = spectrum_query.attrib['start_scan']
						end_scan = spectrum_query.attrib['end_scan']
						assumed_charge = spectrum_query.attrib['assumed_charge']
						retention_time_sec = spectrum_query.attrib['retention_time_sec']

						ion_mobility = np.nan
						if 'ion_mobility' in spectrum_query.attrib:
							ion_mobility = spectrum_query.attrib['ion_mobility']

						for search_result in spectrum_query.findall(".//pepxml_ns:search_result", namespaces):
							prev_pep = None
							for search_hit in search_result.findall(".//pepxml_ns:search_hit", namespaces):
								hit_rank = search_hit.attrib['hit_rank']
								massdiff = search_hit.attrib['massdiff']

								# parse peptide and protein information
								peptide = search_hit.attrib['peptide']
								unprocessed_proteins = [search_hit.attrib['protein']]

								for alternative_protein in search_hit.findall('.//pepxml_ns:alternative_protein', namespaces):
									unprocessed_proteins.append(alternative_protein.attrib['protein'])

								# remove decoy results from mixed target/decoy hits
								has_targets = False
								has_decoys = False
								for prot in unprocessed_proteins:
									if decoy_prefix in prot:
										has_decoys = True
									else:
										has_targets = True

								processed_proteins = []
								for prot in unprocessed_proteins:
									if has_targets and has_decoys:
										if decoy_prefix not in prot:
											processed_proteins.append(prot)
									else:
										processed_proteins.append(prot)
								num_tot_proteins = len(processed_proteins)

								is_decoy = False
								if has_decoys and not has_targets:
									is_decoy = True

								proteins = {}
								for prot in processed_proteins:
									# Remove UniProt prefixes if necessary
									if decoy_prefix + "sp|" in prot:
										proteins[decoy_prefix + prot.split("|")[1]] = ""
									elif "sp|" in prot:
										proteins[prot.split("|")[1]] = prot.split("|")[2].split(" ")[0].split("_")[0]
									else:
										proteins[prot] = prot

								protein = ""
								gene = ""

								for key in sorted(proteins):
									if protein == "":
										protein = key
									else:
										protein = protein + ";" + key

									if gene == "":
										gene = proteins[key]
									else:
										gene = gene + ";" + proteins[key]

								# parse PTM information
								modifications = "M"
								nterm_modification = ""
								cterm_modification = ""
								for modification_info in search_hit.findall('.//pepxml_ns:modification_info', namespaces):
									if 'mod_nterm_mass' in modification_info.attrib:
										nterm_modification = float(modification_info.attrib['mod_nterm_mass'])
									if 'mod_cterm_mass' in modification_info.attrib:
										cterm_modification = float(modification_info.attrib['mod_cterm_mass'])
									for mod_aminoacid_mass in modification_info.findall('.//pepxml_ns:mod_aminoacid_mass', namespaces):
										modifications = modifications + "|" + mod_aminoacid_mass.attrib['position'] + "$" + mod_aminoacid_mass.attrib['mass']

								# parse search engine score information
								scores = {}
								for search_score in search_hit.findall('.//pepxml_ns:search_score', namespaces):
									scores["var_" + search_score.attrib['name']] = float(search_score.attrib['value'])

								# parse PeptideProphet or iProphet results if available
								for analysis_result in search_hit.findall('.//pepxml_ns:analysis_result', namespaces):
									if analysis_result.attrib['analysis'] == 'interprophet':
										for interprophet_result in analysis_result.findall('.//pepxml_ns:interprophet_result', namespaces):
											scores["pep"] = 1.0 - float(interprophet_result.attrib['probability'])
											prev_pep = scores["pep"]
									elif analysis_result.attrib['analysis'] == 'peptideprophet':
										for peptideprophet_result in analysis_result.findall('.//pepxml_ns:peptideprophet_result', namespaces):
											scores["pep"] = 1.0 - float(peptideprophet_result.attrib['probability'])
											prev_pep = scores["pep"]

								if prev_pep is not None and "pep" not in scores:
									# If 2 search hits have the same rank only the first one has the analysis_result explicitly written out.
									scores["pep"] = prev_pep

								if prev_pep is not None:
									peptides.append({**{'run_id': base_name, 'scan_id': int(start_scan), 'hit_rank': int(hit_rank), 'massdiff': float(massdiff), 'precursor_charge': int(assumed_charge), 'retention_time': float(retention_time_sec), 'ion_mobility': float(ion_mobility), 'peptide_sequence': peptide, 'modifications': modifications, 'nterm_modification': nterm_modification, 'cterm_modification': cterm_modification, 'protein_id': protein, 'gene_id': gene, 'num_tot_proteins': num_tot_proteins, 'decoy': is_decoy}, **scores})
				elem.clear()

		df = pd.DataFrame(peptides)

		return(df)

class idxml:
    def __init__(self, idxml_file, base_name):
        self.idxml_file = idxml_file
        self.base_name = base_name
        self.psms = self.parse_idxml()
        #self.exclude_range = exclude_range
        #self.match_unimod(unimod)

    def get(self):
        return(self.psms)

    def parse_idxml(self):
        peptides = []
        proteins = []
        scores = {}

        parsed_peptides = []

        po.IdXMLFile().load(self.idxml_file, proteins, peptides)

        for p in peptides:

            #percolator probability
            scores["q_value"] = float(p.getHits()[0].getMetaValue('MS:1001491'))
            scores["pep"] = float(p.getHits()[0].getMetaValue('MS:1001491'))

            try:
                parsed_peptides.append({**{'run_id': self.base_name,
                                           'scan_id': int(str(p.getMetaValue("spectrum_reference")).split('scan=')[-1].strip("'")),
                                           'hit_rank': int(p.getHits()[0].getRank()),
                                           'massdiff': float(0),
                                           'precursor_charge': int(p.getHits()[0].getCharge()),
                                           'retention_time': float(p.getRT()),
                                           'modified_peptide': p.getHits()[0].getSequence().toUniModString().decode("utf-8"),
                                           'peptide_sequence': p.getHits()[0].getSequence().toUnmodifiedString().decode("utf-8"),
                                           'modifications': '-',
                                           'nterm_modification': '-',
                                           'cterm_modification': '-',
                                           'protein_id': ','.join([prot.getProteinAccession().decode("utf-8") for prot in p.getHits()[0].getPeptideEvidences()]),
                                           'gene_id': '-',
                                           'num_tot_proteins': len([prot.getProteinAccession() for prot in p.getHits()[0].getPeptideEvidences()]),
                                           'decoy': p.getHits()[0].getMetaValue('target_decoy').decode("utf-8")=='decoy'}, **scores})

            except AttributeError:
                parsed_peptides.append({**{'run_id': self.base_name,
                                           'scan_id': int(str(p.getMetaValue("spectrum_reference")).split('scan=')[-1].strip("'")),
                                           'hit_rank': int(p.getHits()[0].getRank()),
                                           'massdiff': float(0),
                                           'precursor_charge': int(p.getHits()[0].getCharge()),
                                           'retention_time': float(p.getRT()),
                                           'modified_peptide': p.getHits()[0].getSequence().toUniModString(),
                                           'peptide_sequence': p.getHits()[0].getSequence().toUnmodifiedString(),
                                           'modifications': '-',
                                           'nterm_modification': '-',
                                           'cterm_modification': '-',
                                           'protein_id': ','.join([prot.getProteinAccession() for prot in p.getHits()[0].getPeptideEvidences()]),
                                           'gene_id': '-',
                                           'num_tot_proteins': len([prot.getProteinAccession() for prot in p.getHits()[0].getPeptideEvidences()]),
                                           'decoy': p.getHits()[0].getMetaValue('target_decoy')=='decoy'}, **scores})

        df = pd.DataFrame(parsed_peptides)

        return (df)

class unimod:
	def __init__(self, unimod_file, max_delta):
		self.unimod_file = unimod_file
		self.max_delta = max_delta
		self.ptms = self.parse_unimod()

	def parse_unimod(self):
		namespaces = {'umod': "http://www.unimod.org/xmlns/schema/unimod_2"}
		ET.register_namespace('', "http://www.unimod.org/xmlns/schema/unimod_2")
		tree = ET.parse(self.unimod_file)
		root = tree.getroot()

		ptms = {}
		sites = ['A','R','N','D','C','E','Q','G','H','O','I','L','K','M','F','P','U','S','T','W','Y','V','N-term','C-term', 'B', 'J', 'X', 'Z']
		positions = ['Anywhere','Any N-term','Any C-term','Protein N-term','Protein C-term']

		for site in sites:
			ptms[site] = {}

		for position in positions:
			for site in sites:
				ptms[site][position] = {}

		for modifications in root.findall('.//umod:modifications', namespaces):
			for modification in modifications.findall('.//umod:mod', namespaces):
				for specificity in modification.findall('.//umod:specificity', namespaces):
					ptms[specificity.attrib['site']][specificity.attrib['position']][int(modification.attrib['record_id'])] = float(modification.findall('.//umod:delta', namespaces)[0].attrib['mono_mass'])

		return ptms

	def get_id(self, site, position, delta_mass):
		candidates = {}
		min_id = -1
		ptms_site = self.ptms[site]
		search_multiple_positions = isinstance(position, (list, tuple))

		kvs = itertools.chain.from_iterable((((k, p), v) for k,v in ptms_site[p].items()) for p in position) \
			if search_multiple_positions else \
			ptms_site[position].items()
		for key, value in kvs:
			delta_mod = abs(value - float(delta_mass))
			if delta_mod < self.max_delta:
				candidates_key = candidates.setdefault(key, delta_mod)
				if delta_mod < candidates_key:
					candidates[key] = delta_mod

		if len(candidates) > 0:
			min_id = min(candidates, key=candidates.get)

		return(min_id)

	def get_oms_id(self, sequence, massdiff, nterm_modification, cterm_modification):
		record_ids = {}
		for site, aa in enumerate(sequence):
			if aa != "_":
				record_id_site = self.get_id(aa, 'Anywhere', massdiff)
				if record_id_site != -1:
					record_ids[site+1] = record_id_site

		record_id_nterm = -1
		if nterm_modification == "":
			record_id_nterm = self.get_id("N-term", 'Any N-term', massdiff)
			if record_id_nterm == -1:
				record_id_nterm = self.get_id("N-term", 'Protein N-term', massdiff)

		record_id_cterm = -1
		if cterm_modification == "":
			record_id_cterm = self.get_id("C-term", 'Any C-term', massdiff)
			if record_id_cterm == -1:
				record_id_cterm = self.get_id("C-term", 'Protein C-term', massdiff)

		# prefer residual over N-term over C-term modifications
		aamod = {}
		if len(record_ids) > 0:
			aasite = median_low(list(record_ids.keys()))
			aamod[aasite] = massdiff
		elif record_id_nterm != -1:
			nterm_modification = massdiff
		elif record_id_cterm != -1:
			cterm_modification = massdiff

		return aamod, nterm_modification, cterm_modification


good_patterns = [re.compile('controllerType=0 controllerNumber=1 scan=([0-9]+)'),
				re.compile('frame=([0-9]+)'),
				re.compile('function=[0-9]+ process=[0-9]+ scan=([0-9]+)'),
				re.compile('jobRun=[0-9]+ spotLabel=[^ ]+ spectrum=([0-9]+)'),
				re.compile('([0-9]+)'),
				re.compile('scan=([0-9]+)'),
				re.compile('spectrum=([0-9]+)'),
				re.compile('scanId=([0-9]+)'),
				re.compile('index=([0-9]+)')]

bad_patterns = [re.compile('controllerType=[1-9] controllerNumber=1 scan=[0-9]+'),
				re.compile('controllerType=[0-9] controllerNumber=1 scan=[0-9]+ demux=[0-9]+')]


def get_scan(e: str, fallback_num: int):
	if e is None or e == "":
		return fallback_num

	for bad_pattern in bad_patterns:
		if bad_pattern.fullmatch(e) is not None:
			return fallback_num

	for good_pattern in good_patterns:
		x = good_pattern.fullmatch(e)
		if x is not None:
			return int(x.group(1))

	return fallback_num


def read_mzml_or_mzxml_impl(input_map, psms, theoretical, max_delta_ppm, filetype):
	assert filetype in ('mzml', 'mzxml')

	timestamped_echo("Info: Collecting PSMs...")

	input_map = {get_scan(getNativeID, idx + 1): e for idx, (getNativeID, e) in enumerate(input_map)}
	import concurrent.futures
	nthreads = min(os.cpu_count(), 5)
	def f(psms):
		peaks_list = []
		for scan_id, modified_peptide, precursor_charge in psms.itertuples(index=None):
			peaks_list.append(
				psm_df(input_map, theoretical, max_delta_ppm, scan_id, modified_peptide, precursor_charge))
		return peaks_list
	with concurrent.futures.ThreadPoolExecutor(nthreads) as exe:
		l = exe.map(f, np.array_split(psms, nthreads))
	peaks_list = sum(l, [])

	timestamped_echo("Info: Got %d PSMs" % len(peaks_list))
	timestamped_echo("Info: Generating Transitions...")

	if len(peaks_list) > 0:
		reps = np.array([e[0] for e in peaks_list])
		transitions = pd.DataFrame({'fragment': np.concatenate([e[1] for e in peaks_list]),
									'product_mz': np.concatenate([e[2] for e in peaks_list]),
									'intensity': np.concatenate([e[3] for e in peaks_list]),
									'scan_id': np.repeat([e[4] for e in peaks_list], reps),
									'precursor_mz': np.repeat([e[5] for e in peaks_list], reps),
									'modified_peptide': np.repeat([e[6] for e in peaks_list], reps),
									'precursor_charge': np.repeat([e[7] for e in peaks_list], reps)})
		# Multiple peaks might be identically annotated, only use most intense
		transitions = transitions.groupby(['scan_id','modified_peptide','precursor_charge','precursor_mz','fragment','product_mz'])['intensity'].max().reset_index()
	else:
		transitions = pd.DataFrame({'scan_id': [], 'modified_peptide': [], 'precursor_charge': [], 'precursor_mz': [], 'fragment': [], 'product_mz': [], 'intensity': []})

	timestamped_echo("Info: Generated %d transitions" % len(transitions))

	return(transitions)


def read_mgf(tims_mgf_path, psms, theoretical, max_delta_ppm):
	'''
	read TimsTOF's MGF or MSFragger's calibrated.mgf
	'''
	is_tims = not tims_mgf_path.endswith('calibrated.mgf')
	import mmap
	record_pattern = re.compile(b'''BEGIN IONS\r?
(.*?)
END IONS''', re.MULTILINE | re.DOTALL)
	scan_num_pattern_Bruker = re.compile(rb'TITLE=Cmpd\s+([0-9]+),')
	scan_num_pattern_general = re.compile(rb'TITLE=.+?\.([0-9]+?)\.')
	peaks_pattern = re.compile(rb'^([\d.]+)\s+([\d.]+)', re.MULTILINE)

	tims_data = {}
	with open(tims_mgf_path, "rb") as f:
		mm = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
		for e in record_pattern.finditer(mm):
			rec = e.group(1)
			scan_num_findall = (scan_num_pattern_Bruker.findall(rec) + scan_num_pattern_general.findall(rec))
			if len(scan_num_findall) in [1, 2]:
				scan_num = int(scan_num_findall[0])
			else:
				raise RuntimeError("Cannot find Cmpd number from " + rec.decode()
								   if is_tims else
								   "Cannot find scan number from " + rec.decode())
			tims_data[scan_num] = np.array(peaks_pattern.findall(rec), dtype=float)

	peaks_list = []
	for scan_id, modified_peptide, precursor_charge in psms.itertuples(index=False):
		peaks_list.append(psm_df_mgf(tims_data, theoretical, max_delta_ppm, scan_id, modified_peptide, precursor_charge))

	if len(peaks_list) > 0:
		reps = np.array([e[0] for e in peaks_list])
		transitions = pd.DataFrame({'fragment': np.concatenate([e[1] for e in peaks_list]),
									'product_mz': np.concatenate([e[2] for e in peaks_list]),
									'intensity': np.concatenate([e[3] for e in peaks_list]),
									'scan_id': np.repeat([e[4] for e in peaks_list], reps),
									'precursor_mz': np.repeat([e[5] for e in peaks_list], reps),
									'modified_peptide': np.repeat([e[6] for e in peaks_list], reps),
									'precursor_charge': np.repeat([e[7] for e in peaks_list], reps)})
		# Multiple peaks might be identically annotated, only use most intense
		transitions = transitions.groupby(['scan_id','modified_peptide','precursor_charge','precursor_mz','fragment','product_mz'])['intensity'].max().reset_index()
	else:
		transitions = pd.DataFrame({'scan_id': [], 'modified_peptide': [], 'precursor_charge': [], 'precursor_mz': [], 'fragment': [], 'product_mz': [], 'intensity': []})
	return transitions


def annotate_mass(mass, ionseries, max_delta_ppm):
	top_delta = 30
	ions, ion_masses = ionseries
	ppms = np.abs((mass - ion_masses) / ion_masses * 1e6)
	idx = ppms.argmin()
	if ppms[idx] < min(max_delta_ppm, top_delta):
		return ions[idx], ion_masses[idx]
	return None, None


def psm_df(input_map, theoretical, max_delta_ppm, scan_id, modified_peptide, precursor_charge):
	ionseries = theoretical[modified_peptide][precursor_charge]

	spectrum = input_map[scan_id]

	fragments, product_mzs, intensities = annotate_mass_spectrum_numba(ionseries, max_delta_ppm, spectrum)
	# Baseline normalization to highest annotated peak
	max_intensity = np.amax(intensities, initial=0.0)
	if max_intensity > 0:
		intensities /= max_intensity
		intensities *= 10000
	return [len(fragments), fragments, product_mzs, intensities,
			scan_id,
			po.AASequence.fromString(po.String(modified_peptide))
				.getMonoWeight(po.Residue.ResidueType.Full, precursor_charge) / precursor_charge,
			modified_peptide, precursor_charge]

def psm_df_mgf(input_map, theoretical, max_delta_ppm, scan_id, modified_peptide, precursor_charge):
	ionseries = theoretical[modified_peptide][precursor_charge]

	spectrum = input_map[scan_id]

	top_delta = 30
	ions, ion_masses = ionseries

	mzs0, intensities0 = spectrum[:, 0], spectrum[:, 1]
	ppms = np.abs((mzs0[:, np.newaxis] - ion_masses) / ion_masses * 1e6)
	idx_mask = (ppms < min(max_delta_ppm, top_delta)).any(1)
	idx = ppms[idx_mask].argmin(1)
	fragments, product_mzs, intensities = ions[idx], ion_masses[idx], intensities0[idx_mask]

	# Baseline normalization to highest annotated peak
	max_intensity = np.amax(intensities, initial=0.0)
	if max_intensity > 0:
		intensities /= max_intensity
		intensities *= 10000
	return [len(fragments), fragments, product_mzs, intensities,
			scan_id,
			po.AASequence.fromString(po.String(modified_peptide))
				.getMonoWeight(po.Residue.ResidueType.Full, precursor_charge) / precursor_charge,
			modified_peptide, precursor_charge]


def annotate_mass_spectrum(ionseries, max_delta_ppm, spectrum):
	top_delta = 30
	ions, ion_masses = ionseries

	mzs0, intensities0 = spectrum#.get_peaks()
	ppms = np.abs((mzs0[:, np.newaxis] - ion_masses) / ion_masses * 1e6)
	idx_mask = (ppms < min(max_delta_ppm, top_delta)).any(1)
	idx = ppms[idx_mask].argmin(1)
	return ions[idx], ion_masses[idx], intensities0[idx_mask]

import numba
@numba.jit(nopython=True, nogil=True, fastmath=True)
def annotate_mass_spectrum_numba(ionseries, max_delta_ppm, spectrum):
	top_delta = 30
	ions, ion_masses = ionseries
	mzs0, intensities0 = spectrum
	idx_ions = np.empty_like(intensities0, dtype=np.uint32)
	idx_peaks = np.zeros_like(intensities0, dtype=np.bool_)
	for si, mz in enumerate(mzs0):
		min_ion_idx = -1
		min_ppm = top_delta
		for ii, ion_mass in enumerate(ion_masses):
			ppm = np.abs(mz - ion_mass) / ion_mass * 1e6
			if ppm < min(max_delta_ppm, top_delta):
				if ppm < min_ppm:
					min_ppm = ppm
					min_ion_idx = ii
		if min_ion_idx != -1:
			idx_ions[si] = min_ion_idx
			idx_peaks[si] = True
	return ions[idx_ions[idx_peaks]], ion_masses[idx_ions[idx_peaks]], intensities0[idx_peaks]


def generate_ionseries(peptide_sequence, precursor_charge, fragment_charges=[1,2,3,4], fragment_types=['b','y'], enable_specific_losses = False, enable_unspecific_losses = False, precision_digits = 6):
	peptide = po.AASequence.fromString(po.String(peptide_sequence))
	sequence = peptide.toUnmodifiedString()

	# this dict is compatible with DIA-NN
	unspecific_losses = dict()
	unspecific_losses["H2O"] = 18.0106
	unspecific_losses["NH3"] = 17.0265
	# unspecific_losses["CO"] = 27.9949
	# unspecific_losses["H3PO4"] = 97.9769
	# unspecific_losses["H4COS"] = 63.9983

	fragments = {}

	for fragment_type in fragment_types:
		for fragment_charge in fragment_charges:
			if fragment_charge <= precursor_charge:
				for fragment_ordinal in range(1,len(sequence)):
					mass = ion = None
					if fragment_type == 'a':
						ion = peptide.getPrefix(fragment_ordinal)
						mass = ion.getMonoWeight(po.Residue.ResidueType.AIon, fragment_charge) / fragment_charge;
					elif fragment_type == 'b':
						ion = peptide.getPrefix(fragment_ordinal)
						mass = ion.getMonoWeight(po.Residue.ResidueType.BIon, fragment_charge) / fragment_charge;
					elif fragment_type == 'c':
						ion = peptide.getPrefix(fragment_ordinal)
						mass = ion.getMonoWeight(po.Residue.ResidueType.CIon, fragment_charge) / fragment_charge;
					elif fragment_type == 'x':
						ion = peptide.getSuffix(fragment_ordinal)
						mass = ion.getMonoWeight(po.Residue.ResidueType.XIon, fragment_charge) / fragment_charge;
					elif fragment_type == 'y':
						ion = peptide.getSuffix(fragment_ordinal)
						mass = ion.getMonoWeight(po.Residue.ResidueType.YIon, fragment_charge) / fragment_charge;
					elif fragment_type == 'z':
						ion = peptide.getSuffix(fragment_ordinal)
						mass = ion.getMonoWeight(po.Residue.ResidueType.ZIon, fragment_charge) / fragment_charge;
					else:
						raise RuntimeError(f'fragment type "{fragment_type}" is not in (a,b,c,x,y,z)')
					
					# Workaround
					# If two fragment ions have identical product m/z, this can lead to annotation inconsistencies.
					# E.g. .(UniMod:1)ADQLTEEQIAEFK+2 and the corresponding b5^1 and b10^2 ions.
					# We thus first generate a dict with product m/z as index and the reserve the dict.
					# Then do not update the dicts if the product m/z is already present.
					# This leads to only the lower charged product to be annotated and reported.

					# Standard fragment ions
					k = round(mass, precision_digits)
					if k not in fragments:
						fragments[k] = fragment_type + str(fragment_ordinal) + "^" + str(fragment_charge)

					# unspecific losses that are compatible with DIA-NN
					if enable_unspecific_losses:
						for loss in unspecific_losses:
							k = round(mass - (unspecific_losses[loss] / fragment_charge), precision_digits)
							if k not in fragments:
								fragments[k] = fragment_type + str(fragment_ordinal) + "-" + loss + "^" + str(fragment_charge)

					# specific losses that are hardcoded in OpenMS
					if enable_specific_losses:
						for lossfragment_ordinal in range(1, ion.size()):
							if ion.getResidue(lossfragment_ordinal).hasNeutralLoss():
								losses = ion.getResidue(lossfragment_ordinal).getLossFormulas()
								for loss in losses:
									loss_type = loss.toString()
									if loss_type not in unspecific_losses:
										k = round(mass - (loss.getMonoWeight() / fragment_charge), precision_digits)
										fragments[k] = fragment_type + str(fragment_ordinal) + "-" + loss_type + "^" + str(fragment_charge)

	# flip key - values for workaround
	fragments = {value: key for key, value in fragments.items()}
	return np.array(list(fragments.keys())), np.fromiter(fragments.values(), float, len(fragments))

def parse_pepxmls(pepxmlfile_list, um, base_name, exclude_range, enable_unannotated, enable_massdiff, fragment_charges, fragment_types, enable_specific_losses, enable_unspecific_losses, precision_digits):
	psmslist = []
	for pepxmlfile in pepxmlfile_list:
		if pepxmlfile.casefold().endswith(('.pepxml', '.pep.xml')):
			timestamped_echo(f"Info: Parsing pepXML: {pepxmlfile}")
			px = pepxml(pepxmlfile, um, base_name, exclude_range, enable_unannotated, enable_massdiff)
		elif pepxmlfile.lower().endswith('idxml'):
			timestamped_echo(f"Info: Parsing idXML: {pepxmlfile}")
			px = idxml(pepxmlfile, base_name)
		else:
			timestamped_echo('unknown format of pepxml identification file')
		psms = px.get()
		if psms.shape[0] > 0:
			rank = re.compile(r'_rank([0-9]+)\.').search(pathlib.Path(pepxmlfile).name)
			rank_str = '' if rank is None else '_rank' + rank.group(1)
			psms['group_id'] = psms['run_id'] + "_" + psms['scan_id'].astype(str) + rank_str
			psmslist.append(psms)
		timestamped_echo(f"Info: {psms.shape[0]} PSMs parsed.")

	out_df = pd.DataFrame()
	theoretical = None
	if len(psmslist) > 0:
		out_df = pd.concat(psmslist)
		if out_df.shape[0] > 0:
			timestamped_echo("Info: Generate theoretical spectra.")
			theoretical = {}
			for modified_peptide, precursor_charge in out_df[['modified_peptide','precursor_charge']].drop_duplicates().itertuples(index=False):
				theoretical.setdefault(modified_peptide, {})[precursor_charge] = generate_ionseries(modified_peptide, precursor_charge, fragment_charges, fragment_types, enable_specific_losses, enable_unspecific_losses, precision_digits)
	return out_df, theoretical

class MSCallback:
	def __init__(self):
		self.id_peaks_map = []

	def setExperimentalSettings(self, s):
		pass

	def setExpectedSize(self, a, b):
		pass

	def consumeChromatogram(self, c):
		pass
	def consumeSpectrum(self, s):
		self.id_peaks_map.append((s.getNativeID(), s.get_peaks()))

def parse_psms(psm_file_list, um, base_name, exclude_range, enable_unannotated, ignore_unannotated, enable_massdiff, fragment_charges, fragment_types, enable_specific_losses, enable_unspecific_losses, decoy_prefix, precision_digits, labile_mods, max_glycan_q):
	"""
	Parsing method with psm.tsv as the primary input instead of pepxml/idxml
	"""
	psmslist = []
	for psmfile in psm_file_list:
		px = psmtsv(psmfile, um, base_name, exclude_range, enable_unannotated, ignore_unannotated, enable_massdiff, decoy_prefix, labile_mods, max_glycan_q)
		psms = px.get()
		psms['group_id'] = psms['run_id'] + "_" + psms['scan_id'].astype(str) + "_rank" + psms['hit_rank'].astype(str)
		click.echo(f"Info: Done parsing psm.tsv: {psmfile}")
		psmslist.append(psms)
	psms = pd.concat(psmslist)
	theoretical = None
	if psms.shape[0] > 0:
		# Generate theoretical spectra
		click.echo("Info: Generate theoretical spectra.")
		theoretical = {}
		if labile_mods != '':
			for modified_peptide, precursor_charge, labile_peptide in psms[['modified_peptide', 'precursor_charge', 'labile_modified_peptide']].drop_duplicates().itertuples(index=False):
				theoretical.setdefault(modified_peptide, {})[precursor_charge] = generate_ionseries(labile_peptide, precursor_charge, fragment_charges, fragment_types, enable_specific_losses, enable_unspecific_losses, precision_digits)
		else:
			for modified_peptide, precursor_charge in psms[['modified_peptide', 'precursor_charge']].drop_duplicates().itertuples(index=False):
				theoretical.setdefault(modified_peptide, {})[precursor_charge] = generate_ionseries(modified_peptide, precursor_charge, fragment_charges, fragment_types, enable_specific_losses, enable_unspecific_losses, precision_digits)
	return psms, theoretical


def get_map_mzml_or_mzxml(path: str, filetype):
	assert filetype in ('mzml', 'mzxml')
	fh = po.MzMLFile() if filetype=='mzml' else po.MzXMLFile()
	consumer = MSCallback()
	fh.transform(path, consumer)
	return consumer.id_peaks_map

def conversion(pepxmlfile_list, spectralfile, unimodfile, exclude_range, max_delta_unimod, max_delta_ppm, enable_unannotated, enable_massdiff, fragment_types, fragment_charges, enable_specific_losses, enable_unspecific_losses, max_psm_pep, precision_digits):
	# Parse basename
	base_name = basename_spectralfile(spectralfile)
	timestamped_echo("Info: Parsing run %s." % base_name)

	# Initialize UniMod
	um = unimod(unimodfile, max_delta_unimod)

	psms, theoretical = parse_pepxmls(pepxmlfile_list, um, base_name, exclude_range, enable_unannotated, enable_massdiff, fragment_charges, fragment_types, enable_specific_losses, enable_unspecific_losses, precision_digits)

	timestamped_echo("Info: Processing spectra from file %s." % spectralfile)

	if spectralfile.lower().endswith(".mzxml"):
		input_map = get_map_mzml_or_mzxml(spectralfile, 'mzxml')
	elif spectralfile.casefold().endswith(".mzml"):
		input_map = get_map_mzml_or_mzxml(spectralfile, 'mzml')
	else:
		input_map = None

	timestamped_echo("Info: Loaded %d spectra" % len(input_map))

	if psms.shape[0] > 0:
		# Generate spectrum dataframe
		psms = psms[psms['pep'] <= max_psm_pep]
		if spectralfile.lower().endswith(".mzxml"):
			peaks = read_mzml_or_mzxml_impl(input_map, psms[['scan_id','modified_peptide','precursor_charge']], theoretical, max_delta_ppm, 'mzxml')
		elif spectralfile.casefold().endswith(".mzml"):
			peaks = read_mzml_or_mzxml_impl(input_map, psms[['scan_id', 'modified_peptide', 'precursor_charge']], theoretical, max_delta_ppm, 'mzml')
		elif spectralfile.lower().endswith(".mgf"):
			peaks = read_mgf(spectralfile, psms[['scan_id', 'modified_peptide', 'precursor_charge']], theoretical, max_delta_ppm)

		# Round floating numbers
		peaks = peaks.round(6)

		return psms, peaks
	else:
		return pd.DataFrame({'run_id': [], 'scan_id': [], 'hit_rank': [], 'massdiff': [], 'precursor_charge': [], 'retention_time': [], 'ion_mobility': [], 'peptide_sequence': [], 'modifications': [], 'nterm_modification': [], 'cterm_modification': [], 'protein_id': [], 'gene_id': [], 'num_tot_proteins': [], 'decoy': []}), pd.DataFrame({'scan_id': [], 'modified_peptide': [], 'precursor_charge': [], 'precursor_mz': [], 'fragment': [], 'product_mz': [], 'intensity': []})


def conversion_psm(psm_file_list, spectralfile, unimodfile, exclude_range, max_delta_unimod, max_delta_ppm, enable_unannotated, ignore_unannotated, enable_massdiff, fragment_types, fragment_charges, enable_specific_losses, enable_unspecific_losses, max_psm_pep, decoy_prefix, precision_digits, labile_mods, max_glycan_q):
	# Parse basename
	base_name = basename_spectralfile(spectralfile)
	timestamped_echo("Info: Parsing run %s." % base_name)

	# Initialize UniMod
	um = unimod(unimodfile, max_delta_unimod)
	import concurrent.futures

	timestamped_echo("Info: Processing spectra from file %s." % spectralfile)

	exe = concurrent.futures.ProcessPoolExecutor(1)
	psms_fut = exe.submit(parse_psms, psm_file_list, um, base_name, exclude_range, enable_unannotated, ignore_unannotated, enable_massdiff, fragment_charges, fragment_types, enable_specific_losses, enable_unspecific_losses, decoy_prefix, precision_digits, labile_mods, max_glycan_q)
	time.sleep(1)  # allow the process to execute first before using pyOpenMS to read files
	if spectralfile.lower().endswith(".mzxml"):
		input_map = get_map_mzml_or_mzxml(spectralfile, 'mzxml')
	elif spectralfile.casefold().endswith(".mzml"):
		input_map = get_map_mzml_or_mzxml(spectralfile, 'mzml')
	else:
		input_map = None

	timestamped_echo("Info: Loaded %d spectra" % len(input_map))
	timestamped_echo("Info: Processing PSM file...")

	# Continue if any PSMS are present
	psms, theoretical = psms_fut.result()
	exe.shutdown()

	if psms.shape[0] > 0:
		# Generate spectrum dataframe
		click.echo("Info: Processing spectra from file %s." % spectralfile)
		psms = psms[psms['pep'] <= max_psm_pep]
		if spectralfile.lower().endswith(".mzxml"):
			peaks = read_mzml_or_mzxml_impl(input_map, psms[['scan_id','modified_peptide','precursor_charge']], theoretical, max_delta_ppm, 'mzxml')
		elif spectralfile.casefold().endswith(".mzml"):
			peaks = read_mzml_or_mzxml_impl(input_map, psms[['scan_id', 'modified_peptide', 'precursor_charge']], theoretical, max_delta_ppm, 'mzml')
		elif spectralfile.lower().endswith(".mgf"):
			peaks = read_mgf(spectralfile, psms[['scan_id', 'modified_peptide', 'precursor_charge']], theoretical, max_delta_ppm)

		# Round floating numbers
		peaks = peaks.round(6)

		return psms, peaks
	else:
		return pd.DataFrame({'run_id': [], 'scan_id': [], 'hit_rank': [], 'massdiff': [], 'precursor_charge': [], 'retention_time': [], 'ion_mobility': [], 'peptide_sequence': [], 'modifications': [], 'nterm_modification': [], 'cterm_modification': [], 'protein_id': [], 'gene_id': [], 'num_tot_proteins': [], 'decoy': []}), pd.DataFrame({'scan_id': [], 'modified_peptide': [], 'precursor_charge': [], 'precursor_mz': [], 'fragment': [], 'product_mz': [], 'intensity': []})


def basename_spectralfile(spectralfile):
	'''
	take the basename of a spectral filename
	and strip trailing `_calibrated` if any, for MSFragger's MGF.
	:param spectralfile: name of spectral file
	:return: basename without trailing `_calibrated`
	'''
	x = os.path.splitext(os.path.basename(spectralfile))[0]
	# get basename without _(un)calibrated suffix, if any
	return re.compile('(.+?)(?:_(?:un)?calibrated)?').fullmatch(x)[1]



import unittest
from easypqp.convert import get_scan


class TestConvert(unittest.TestCase):

    def test_get_scan(self):
        self.assertEqual(11, get_scan("controllerType=0 controllerNumber=1 scan=2 demux=0", 11))
        self.assertEqual(11, get_scan("sample=2 period=3 cycle=4 experiment=5", 11))
        self.assertEqual(11, get_scan("frame=2 scan=3", 11))

        self.assertEqual(11, get_scan("controllerType=0 controllerNumber=1 scan=11", 22))
        self.assertEqual(11, get_scan("function=0 process=1 scan=11", 22))
        self.assertEqual(11, get_scan("jobRun=0 spotLabel=asw spectrum=11", 22))
        self.assertEqual(11, get_scan("11", 22))
        self.assertEqual(11, get_scan("scan=11", 22))
        self.assertEqual(11, get_scan("spectrum=11", 22))
        self.assertEqual(11, get_scan("scanId=11", 22))
        self.assertEqual(11, get_scan("index=11", 22))
        self.assertEqual(11, get_scan("frame=11", 22))


if __name__ == '__main__':
    unittest.main()






from setuptools import setup, find_packages

from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='easypqp',
    version='0.1.49',
    description='EasyPQP: Simple library generation for OpenSWATH',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/grosenberger/easypqp",
	author="George Rosenberger",
	author_email="gr2578@cumc.columbia.edu",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    include_package_data=True,
    install_requires=['numba','Click>=8.0.0','numpy==1.26.4','scipy','scikit-learn','statsmodels','pandas>=1.5.0','biopython','pyopenms>=3.0.0','matplotlib>=3.5.1','seaborn', 'tqdm'],
    # extras_require={  # Optional
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },
    package_data={  # Optional
        'easypqp': ['data/unimod.xml'],
    },
    entry_points={
        'console_scripts': [
            'easypqp=easypqp.main:cli',
        ],
    },
    extras_require={
        'PyProphet': ['pyprophet'],
    },
)



import click
from tqdm import tqdm

# Unimod parsing
import xml.etree.cElementTree as ET


def site_validation(site_input):
    """
    Perform a check to ensure inputs are valid
    Arguments:
        site_input: (list) list of amino acid residues, or terminal notation, or wild card notation (*).
    Returns:
        Nothing is returned. An error is raised if the input contains a non-valid site.
    """
    acceptable_sites = ["A", "R", "N", "D", "C", "E", "Q", "G", "H", "I", "L", "K", "M", "F", "P", "S", "T", "W", "Y", "V", 'U', 'O', '[', ']', 'n', 'c', '*']
    site_check = [site not in acceptable_sites for site in site_input]
    if any(site_check):
        raise click.ClickException( f"Incorrect site specificity input, site(s) {', '.join([i for (i, v) in zip(site_input, site_check) if v])} is not valid. Acceptable sites: {', '.join(acceptable_sites)}")

def site_specificity_transform(site_input):
    """
    Transform input site to return the site and position. Transforms terminal notation to site notation in unimod.xml and whether its any terminal site or a protein terminal site.
    Arguments:
        site_input: (list) list of amino acid residues, or terminal notation, or wild card notation (*).
    Returns:
        Returns a tuple of list of sites and list of positions
    """
    # Site and Position Mapping
    terminal_map = {'[':'N-term', ']':'C-term', 'n':'N-term', 'c':'C-term'}
    site_position_map = {'[':'Protein N-term', ']':'Protein C-term', 'n':'Any N-term', 'c':'Any C-term'}
    # Split sites
    site_input = [site for site in site_input]
    site_validation(site_input)
    sites=[]; positions=[]
    for site in site_input:
        if site in terminal_map.keys():
            sites.append(terminal_map[site])
            positions.append(site_position_map[site])
        elif site=="*":
            sites.append("*")
            positions.append("*")
        else:
            sites.append(site)
            positions.append("Anywhere")
    return sites, positions

def unimod_filter(unimod_file, out_file, accession_ids, site_specificity):
    """
    Filter an input unimod to restrict for specific modifications and site specificities
    Arguments:
        unimod_file: (str) path/filename of input unimod.xml file.
        out_file: (str) path/filename to write out new filtered unimod.xml file
        accession_ids: (list) list of unimod accession ids to restrict for. i.e. ['1','21','35]
        site_specificity: (list) list of site specificties to further restrict corresponding unimod for. i.e. ['n','STY','M], will restrict acetylation for any N-Term, phosphorylation for serine, threonine, and tyrosine, and oxidation for methionine.
    Returns:
        Nothing is returned. The restricted unimod database is written to the out_file.
    """
    # Register Namespace
    ET.register_namespace('umod', 'http://www.unimod.org/xmlns/schema/unimod_2')

    # Read in unimod XML database
    click.echo(f"INFO: Loading XML data from {unimod_file}")
    tree = ET.parse(unimod_file)
    root = tree.getroot()

    # Namespace
    ns = {'umod':'http://www.unimod.org/xmlns/schema/unimod_2'}

    # Generate root for new filtered unimod XML
    root_out = ET.Element(root.tag, root.attrib)

    # Append elements subelements
    umod_elements = root.findall("umod:elements", ns)
    root_out.append(umod_elements[0])

    # Append desired modifications
    mod_entries = root.findall('umod:modifications', ns)[0]
    mod_out = ET.Element(mod_entries.tag, mod_entries.attrib)
    i=0
    pbar = tqdm(accession_ids)
    pbar_desc = "INFO: Restricting"
    for record_id in pbar:
        add_unimod_entry = mod_entries.findall(f"./umod:mod/[@record_id='{record_id}']", ns)[0]
        if site_specificity is not None:
            site, position = site_specificity_transform(site_specificity[i])
            # Update progess bar description
            pbar_desc = f"INFO: Restricting..{add_unimod_entry.attrib.get('title')}({','.join(site)})"
            pbar.set_description(pbar_desc)
            if site != "*":
                for unimod_site in add_unimod_entry.findall(f"./umod:specificity", ns):
                    if unimod_site.attrib['site'] in site and unimod_site.attrib['position'] in position:
                        # If current specificity element is a requested one, continue on
                        continue
                    else:
                        # Remove specificities that do not match requested specificities
                        add_unimod_entry.remove(unimod_site)
        else:
            # Update progess bar description
            pbar_desc = f"INFO: Restricting..{add_unimod_entry.attrib.get('title')}"
            pbar.set_description(pbar_desc)
        # click.echo(f"INFO: Appending to filtered unimod XML - title={add_unimod_entry.attrib.get('title')} with record_id={add_unimod_entry.attrib.get('record_id')}")
        mod_out.append( add_unimod_entry )
        i+=1
    root_out.append(mod_out)

    # Append amino acids
    umod_amino_acids = root.findall("umod:amino_acids", ns)
    root_out.append(umod_amino_acids[0])

    # Append mod bricks
    umod_mod_bricks = root.findall("umod:mod_bricks", ns)
    root_out.append(umod_mod_bricks[0])

    # Generate element hierarchy to write out to xml
    tree_out = ET.ElementTree(root_out)
    # For Pretty-Printing
    ET.indent(tree_out, '  ')
    # Write out filtered unimod xml database
    click.echo(f"INFO: Writing out filtered unimod XML file to {out_file}")
    tree_out.write(out_file, encoding="UTF-8", xml_declaration=True, method="xml")

    # Insert Top Comment
    # TODO: This may not be the best way to add the top level comment in standard unimod.xml database files. Might be able to use lxml instead, requiring an additional dependency
    with open(out_file, 'r+', encoding="utf-8") as file_handle:
        lines = file_handle.readlines()     
        lines.insert(1, "<!--Copyright (C) 2002-2006 Unimod; this information may be copied, distributed and/or-->\n<!--modified under certain conditions, but it comes WITHOUT ANY WARRANTY; see the-->\n<!--accompanying Design Science License for more details-->\n")  # you can use any index if you know the line index
        file_handle.seek(0)                 
        file_handle.writelines(lines)       


import ast
from .util import timestamped_echo
import time
import importlib.resources as pkg_resources
import click
import sqlite3
import pandas as pd
from shutil import copyfile
from .convert import conversion, basename_spectralfile, conversion_psm
from .library import generate
from .unimoddb import unimod_filter
from easypqp import pkg_unimod_db
from .targetedfileconverter import TargetedFileConverter
from .openswathassaygenerator import OpenSwathAssayGenerator
from .openswathdecoygenerator import OpenSwathDecoyGenerator

try:
    from pyprophet.data_handling import transform_pi0_lambda
except ModuleNotFoundError:
    transform_pi0_lambda = None

@click.group(chain=True)
@click.version_option()
def cli():
    """
    EasyPQP: Simple library generation for OpenSWATH using MSFragger & PyProphet

    Visit https://www.openswath.org for usage instructions and help.
    """


# https://stackoverflow.com/a/47730333
class PythonLiteralOption(click.Option):
    def type_cast_value(self, ctx, value):
        if not isinstance(value, str):  # required for Click>=8.0.0
            return value
        try:
            return ast.literal_eval(value)
        except Exception:
            raise click.BadParameter(value)


# EasyPQP Convert
@cli.command()
@click.option('--pepxml', 'pepxmlfile', required=True, type=click.Path(exists=False), help='The input interact-*.pep.xml file or a list of interact-*.pep.xml files.')
@click.option('--spectra', 'spectralfile', required=True, type=click.Path(exists=True), help='The input mzXML or MGF (timsTOF only) file.')
@click.option('--unimod', 'unimodfile', required=False, type=click.Path(exists=True), help='The input UniMod XML file.')
@click.option('--psms', 'psmsfile', required=False, type=click.Path(exists=False), help='Output PSMs file.')
@click.option('--peaks', 'peaksfile', required=False, type=click.Path(exists=False), help='Output peaks file.')
@click.option('--exclude-range', 'exclude_range_str', default="-1.5,3.5", show_default=True, required=False, type=str, help='massdiff in this range will not be mapped to UniMod.')
@click.option('--max_delta_unimod', default=0.02, show_default=True, type=float, help='Maximum delta mass (Dalton) for UniMod annotation.')
@click.option('--max_delta_ppm', default=15, show_default=True, type=float, help='Maximum delta mass (PPM) for annotation.')
@click.option('--enable_unannotated/--no-enable_unannotated', default=False, show_default=True, help='Enable mapping of unannotated delta masses.')
@click.option('--enable_massdiff/--no-enable_massdiff', default=False, show_default=True, help='Enable mapping of mass differences reported by legacy search engines.')
@click.option('--fragment_types', default="['b','y']", show_default=True, cls=PythonLiteralOption, help='Allowed fragment ion types (a,b,c,x,y,z).')
@click.option('--fragment_charges', default='[1,2,3,4]', show_default=True, cls=PythonLiteralOption, help='Allowed fragment ion charges.')
@click.option('--enable_specific_losses/--no-enable_specific_losses', default=False, show_default=True, help='Enable specific fragment ion losses.')
@click.option('--enable_unspecific_losses/--no-enable_unspecific_losses', default=False, show_default=True, help='Enable unspecific fragment ion losses.')
@click.option('--max_psm_pep', default=0.5, show_default=True, type=float, help='Maximum posterior error probability (PEP) for a PSM')
@click.option('--precision_digits', default=6, show_default=True, type=int, help='Precision (number of digits) for the product m/z reported by the theoretical library generation step. This should match the precision of the downstream consumer of the spectral library. Lowering this number will collapse (more) identical fragment ions of the same precursor to a single value.')
def convert(pepxmlfile, spectralfile, unimodfile, psmsfile, peaksfile, exclude_range_str, max_delta_unimod, max_delta_ppm, enable_unannotated, enable_massdiff, fragment_types, fragment_charges, enable_specific_losses, enable_unspecific_losses, max_psm_pep, precision_digits):
    """
    Convert pepXML files for EasyPQP
    """

    start_time = time.time()

    pepxmlfile_list = []
    if pepxmlfile.endswith(".pep.xml"):
        pepxmlfile_list.append(pepxmlfile)
    elif pepxmlfile.startswith("[") and pepxmlfile.endswith("]"):
        pepxmlfile_list = ast.literal_eval(pepxmlfile)
    else:
        timestamped_echo("Error: Invalid pepXML file name.")
        return 1

    if unimodfile is None:
        unimodfile = str(pkg_resources.files('easypqp').joinpath('data/unimod.xml'))

    run_id = basename_spectralfile(spectralfile)
    if psmsfile is None:
        psmsfile = run_id + ".psmpkl"
    if peaksfile is None:
        peaksfile = run_id + ".peakpkl"

    temp = exclude_range_str.split(',')
    exclude_range = [float(temp[0]), float(temp[1])]

    timestamped_echo("Info: Converting %s." % pepxmlfile_list)
    psms, peaks = conversion(pepxmlfile_list, spectralfile, unimodfile, exclude_range, max_delta_unimod, max_delta_ppm, enable_unannotated, enable_massdiff, fragment_types, fragment_charges, enable_specific_losses, enable_unspecific_losses, max_psm_pep, precision_digits)

    psms.to_pickle(psmsfile)
    timestamped_echo("Info: PSMs successfully converted and stored in %s." % psmsfile)

    peaks.to_pickle(peaksfile)
    timestamped_echo("Info: Peaks successfully converted and stored in %s." % peaksfile)

    timestamped_echo("Info: Total elapsed time %.2f minutes." % ((time.time() - start_time) / 60.0))

# EasyPQP Convert - from psm.tsv rather than pepXML
@cli.command()
@click.option('--psm', 'psmfile', required=True, type=click.Path(exists=False), help='The input psm.tsv file or a list of psm.tsv files.')
@click.option('--spectra', 'spectralfile', required=True, type=click.Path(exists=True), help='The input mzXML or MGF (timsTOF only) file.')
@click.option('--unimod', 'unimodfile', required=False, type=click.Path(exists=True), help='The input UniMod XML file.')
@click.option('--psms', 'psmsfile', required=False, type=click.Path(exists=False), help='Output PSMs file.')
@click.option('--peaks', 'peaksfile', required=False, type=click.Path(exists=False), help='Output peaks file.')
@click.option('--exclude-range', 'exclude_range_str', default="-1.5,3.5", show_default=True, required=False, type=str, help='massdiff in this range will not be mapped to UniMod.')
@click.option('--max_delta_unimod', default=0.02, show_default=True, type=float, help='Maximum delta mass (Dalton) for UniMod annotation.')
@click.option('--max_delta_ppm', default=15, show_default=True, type=float, help='Maximum delta mass (PPM) for annotation.')
@click.option('--enable_unannotated/--no-enable_unannotated', default=False, show_default=True, help='Enable mapping uf unannotated delta masses.')
@click.option('--ignore_unannotated/--no-ignore_unannotated', default=False, show_default=True, help='Remove PSMs with unannotated delta masses but proceed with all other PSMs.')
@click.option('--enable_massdiff/--no-enable_massdiff', default=False, show_default=True, help='Enable mapping uf mass differences reported by legacy search engines.')
@click.option('--fragment_types', default="['b','y']", show_default=True, cls=PythonLiteralOption, help='Allowed fragment ion types (a,b,c,x,y,z).')
@click.option('--fragment_charges', default='[1,2,3,4]', show_default=True, cls=PythonLiteralOption, help='Allowed fragment ion charges.')
@click.option('--enable_specific_losses/--no-enable_specific_losses', default=False, show_default=True, help='Enable specific fragment ion losses.')
@click.option('--enable_unspecific_losses/--no-enable_unspecific_losses', default=False, show_default=True, help='Enable unspecific fragment ion losses.')
@click.option('--max_psm_pep', default=1, show_default=True, type=float, help='Maximum posterior error probability (PEP) for a PSM')
@click.option('--decoy_prefix', default='rev_', show_default=True, type=str, required=True, help='Database decoy prefix (required)')
@click.option('--precision_digits', default=6, show_default=True, type=int, help='Precision (number of digits) for the product m/z reported by the theoretical library generation step. This should match the precision of the downstream consumer of the spectral library. Lowering this number will collapse (more) identical fragment ions of the same precursor to a single value.')
@click.option('--labile_mods', 'labile_mods', default='', show_default=True, required=False, type=str, help='Adjust fragment masses of labile modifications. Supported options: oglyc, nglyc, nglyc+ (includes HexNAc remainder ions)')
@click.option('--max_glycan_q', default=1, show_default=True, type=float, help='Maximum glycan q-value to include a PSM in the library')
def convertpsm(psmfile, spectralfile, unimodfile, psmsfile, peaksfile, exclude_range_str, max_delta_unimod, max_delta_ppm, enable_unannotated, ignore_unannotated, enable_massdiff, fragment_types, fragment_charges, enable_specific_losses, enable_unspecific_losses, max_psm_pep, decoy_prefix, precision_digits, labile_mods, max_glycan_q):
    """
    Convert psm.tsv files for EasyPQP
    """

    start_time = time.time()

    psmfile_list = []
    if psmfile.endswith("psm.tsv"):
        psmfile_list.append(psmfile)
    elif psmfile.startswith("[") and psmfile.endswith("]"):
        psmfile_list = ast.literal_eval(psmfile)
    else:
        timestamped_echo("Error: Invalid psm file name.")
        return 1

    if unimodfile is None:
        unimodfile = str(pkg_resources.files('easypqp').joinpath('data/unimod.xml'))

    run_id = basename_spectralfile(spectralfile)
    if psmsfile is None:
        psmsfile = run_id + ".psmpkl"
    if peaksfile is None:
        peaksfile = run_id + ".peakpkl"

    if labile_mods is not '':
        if labile_mods not in ['oglyc', 'nglyc', 'nglyc+']:
            timestamped_echo("Error: Invalid setting for --labile mods: {}. Allowed options are oglyc, nglyc, or nglyc+".format(labile_mods))
            return 1

    temp = exclude_range_str.split(',')
    exclude_range = [float(temp[0]), float(temp[1])]

    timestamped_echo("Info: Converting %s." % psmfile_list)
    psms, peaks = conversion_psm(psmfile_list, spectralfile, unimodfile, exclude_range, max_delta_unimod, max_delta_ppm, enable_unannotated, ignore_unannotated, enable_massdiff, fragment_types, fragment_charges, enable_specific_losses, enable_unspecific_losses, max_psm_pep, decoy_prefix, precision_digits, labile_mods, max_glycan_q)

    psms.to_pickle(psmsfile)
    timestamped_echo("Info: PSMs successfully converted and stored in %s." % psmsfile)

    peaks.to_pickle(peaksfile)
    timestamped_echo("Info: Peaks successfully converted and stored in %s." % peaksfile)

    timestamped_echo("Info: Total elapsed time %.2f minutes." % ((time.time() - start_time) / 60.0))

# EasyPQP Library
@cli.command()
@click.argument('infiles', nargs=-1, type=click.Path(exists=True))
@click.option('--out', 'outfile', required=True, type=click.Path(exists=False), help='Output TSV peptide query parameter file.')
@click.option('--psmtsv', 'psmtsv', required=False, type=click.Path(exists=False), help='psm.tsv file from Philosopher.')
@click.option('--peptidetsv', 'peptidetsv', required=False, type=click.Path(exists=False), help='peptide.tsv file from Philosopher.')
@click.option('--perform_rt_calibration', required=False, type=bool, help='Whether to perform RT calibration', default=True, show_default=True)
@click.option('--rt_reference', 'rt_referencefile', required=False, type=click.Path(exists=True), help='Optional iRT/CiRT reference file.')
@click.option('--rt_reference_run_path', 'rt_reference_run_path', default='easypqp_rt_reference_run.tsv', show_default=True, required=False, type=click.Path(exists=False), help='Writes reference run RT file, if RT reference file is not provided.')
@click.option('--rt_filter', 'rt_filter', required=False, type=str, help='Optional tag to filter candidate RT reference runs.')
@click.option('--perform_im_calibration', required=False, type=bool, help='Whether to perform IM calibration', default=True, show_default=True)
@click.option('--im_reference', 'im_referencefile', required=False, type=click.Path(exists=True), help='Optional IM reference file.')
@click.option('--im_reference_run_path', 'im_reference_run_path', default='easypqp_im_reference_run.tsv', show_default=True, required=False, type=click.Path(exists=False), help='Writes reference run IM file, if IM reference file is not provided.')
@click.option('--im_filter', 'im_filter', required=False, type=str, help='Optional tag to filter candidate IM reference runs.')
@click.option('--psm_fdr_threshold', default=0.01, show_default=True, type=float, help='PSM FDR threshold.')
@click.option('--peptide_fdr_threshold', default=0.01, show_default=True, type=float, help='Peptide FDR threshold.')
@click.option('--protein_fdr_threshold', default=0.01, show_default=True, type=float, help='Protein FDR threshold.')
@click.option('--rt_lowess_fraction', default=0.05, show_default=True, type=float, help='Fraction of data points to use for RT lowess regression. If set to 0, cross validation is used.')
@click.option('--rt_psm_fdr_threshold', default=0.001, show_default=True, type=float, help='PSM FDR threshold used for RT alignment.')
@click.option('--im_lowess_fraction', default=0.05, show_default=True, type=float, help='Fraction of data points to use for IM lowess regression. If set to 0, cross validation is used.')
@click.option('--im_psm_fdr_threshold', default=0.001, show_default=True, type=float, help='PSM FDR threshold used for IM alignment.')
@click.option('--pi0_lambda', default=[0.1,0.5,0.05], show_default=True, type=(float, float, float), help='Use non-parametric estimation of p-values. Either use <START END STEPS>, e.g. 0.1, 1.0, 0.1 or set to fixed value, e.g. 0.4, 0, 0.', callback=transform_pi0_lambda)
@click.option('--peptide_plot', 'peptide_plot_path', default="easypqp_peptide_report.pdf", show_default=True, required=True, type=click.Path(exists=False), help='Output peptide-level PDF report.')
@click.option('--protein_plot', 'protein_plot_path', default="easypqp_protein_report.pdf", show_default=True, required=True, type=click.Path(exists=False), help='Output protein-level PDF report.')
@click.option('--min_peptides', default=5, show_default=True, type=int, help='Minimum peptides required for successful alignment.')
@click.option('--proteotypic/--no-proteotypic', show_default=True, default=True, help='Use only proteotypic, unique, non-shared peptides.')
@click.option('--consensus/--no-consensus', show_default=True, default=True, help='Generate consensus instead of best replicate spectra.')
@click.option('--nofdr/--no-fdr-filtering', show_default=True, default=False, help='Do not reassess or filter by FDR, as library was already provided using customized FDR filtering.')
def library(infiles, outfile, psmtsv, peptidetsv, perform_rt_calibration, rt_referencefile, rt_reference_run_path, rt_filter, perform_im_calibration, im_referencefile, im_reference_run_path, im_filter, psm_fdr_threshold, peptide_fdr_threshold, protein_fdr_threshold, rt_lowess_fraction, rt_psm_fdr_threshold, im_lowess_fraction, im_psm_fdr_threshold, pi0_lambda, peptide_plot_path, protein_plot_path, min_peptides, proteotypic, consensus, nofdr):
    """
    Generate EasyPQP library
    """

    start_time = time.time()

    generate(infiles, outfile, psmtsv, peptidetsv, perform_rt_calibration, rt_referencefile, rt_reference_run_path, rt_filter, perform_im_calibration, im_referencefile, im_reference_run_path, im_filter, psm_fdr_threshold, peptide_fdr_threshold, protein_fdr_threshold, rt_lowess_fraction, rt_psm_fdr_threshold, im_lowess_fraction, im_psm_fdr_threshold, pi0_lambda, peptide_plot_path, protein_plot_path, min_peptides, proteotypic, consensus, nofdr)
    timestamped_echo("Info: Library successfully generated.")
    timestamped_echo("Info: Total elapsed time: %.2f minutes." % ((time.time() - start_time) / 60.0))

# EasyPQP Reduce
@cli.command()
@click.option('--in', 'infile', required=True, type=click.Path(exists=True), help='Input PQP file.')
@click.option('--out', 'outfile', required=True, type=click.Path(exists=False), help='Output PQP file.')

@click.option('--bins', default=10, show_default=True, type=int, help='Number of bins to fill along gradient.')
@click.option('--peptides', default=5, show_default=True, type=int, help='Number of peptides to sample.')
def reduce(infile, outfile, bins, peptides):
    """
    Reduce PQP files for OpenSWATH linear and non-linear alignment
    """

    # Define outfile
    if outfile is None:
        outfile = infile
    else:
        copyfile(infile, outfile)
        outfile = outfile

    con = sqlite3.connect(outfile)

    anchor_candidates = pd.read_sql('SELECT PRECURSOR.ID AS PRECURSOR_ID, LIBRARY_RT FROM PRECURSOR WHERE PRECURSOR.DECOY == 0;', con)

    anchor_candidates['BIN'] = pd.cut(anchor_candidates['LIBRARY_RT'], bins=bins, right=False, labels=False)

    anchors = anchor_candidates.groupby('BIN').head(peptides).reset_index()

    anchors[['PRECURSOR_ID','LIBRARY_RT']].to_sql('temp_anchors', con, index=False)

    # Delete precursors
    con.execute('DELETE FROM PRECURSOR WHERE ID NOT IN (SELECT PRECURSOR_ID FROM temp_anchors)')

    # Delete transitions
    con.execute('DELETE FROM TRANSITION_PRECURSOR_MAPPING WHERE PRECURSOR_ID NOT IN (SELECT PRECURSOR_ID FROM temp_anchors)')
    con.execute('DELETE FROM TRANSITION WHERE ID NOT IN (SELECT TRANSITION_ID FROM TRANSITION_PRECURSOR_MAPPING)')

    # Delete peptides and proteins
    con.execute('DELETE FROM PRECURSOR_PEPTIDE_MAPPING WHERE PRECURSOR_ID NOT IN (SELECT PRECURSOR_ID FROM temp_anchors)')
    con.execute('DELETE FROM PEPTIDE WHERE ID NOT IN (SELECT PEPTIDE_ID FROM PRECURSOR_PEPTIDE_MAPPING)')
    con.execute('DELETE FROM PEPTIDE_PROTEIN_MAPPING WHERE PEPTIDE_ID NOT IN (SELECT PEPTIDE_ID FROM PRECURSOR_PEPTIDE_MAPPING)')
    con.execute('DELETE FROM PROTEIN WHERE ID NOT IN (SELECT PROTEIN_ID FROM PEPTIDE_PROTEIN_MAPPING)')

    # Delete tables
    con.execute('DROP TABLE temp_anchors;')

    con.commit()

    # Clean file
    con.execute('VACUUM;')

    # Close connection to file
    con.close()

    timestamped_echo("Info: Library successfully processed and stored in %s." % outfile)

# Parameter transformation functions
def transform_comma_string_to_list(ctx, param, value):
    if value is not None:
        str_list = value.split(",")
        return str_list
    else:
        return None 

# EasyPQP UniMod Database Filtering
@cli.command()
@click.option('--in', 'infile', required=False, default=pkg_unimod_db, show_default=True, type=click.Path(exists=True), help='Input UniMod XML file.')
@click.option('--out', 'outfile', required=False, default="unimod_ipf.xml", show_default=True, type=click.Path(exists=False), help='Output Filtered UniMod XML file.')
@click.option('--ids', 'accession_ids', default='1,2,4,5,7,21,26,27,28,34,35,36,40,121,122,259,267,299,354', show_default=True, type=str, help='UniMod record ids to filter for, i.e. 1,2,4,21.', callback=transform_comma_string_to_list)
@click.option('--sites', 'site_specificity', default=None, show_default=True, type=str, help="""Optional further restriction for specificity, i.e. [n,],M,nK[,QN,STY,*,*,*,EDcRK,WM,RK,Y,K,[TKnS,K,R,EK,Y]. Ensure, you match the sites you want to restrict per unimod.\b\n\nFor example, if --ids=1,21,35, then you should have the following for --sites=n,STY,M. This will restrict acetylation for any N-Term, phosphorylation for serine, threonine, and tyrosine, and oxidation for methionine.
\b\n\nValid Sites:\n\n* - wildcard, will not restrict for any specificty for corresponding UniMod entry.\n\n[ - Protein N-Term\n\n] - Protein C-Term\n\nn - Any N-Term\n\nc - Any C-Term\n\nAmino Acid Letter - A valid amino acid one letter code.\n\n""", callback=transform_comma_string_to_list)
def filter_unimod(infile, outfile, accession_ids, site_specificity):
    """
    Reduce UniMod XML Database file
    """
    unimod_filter(infile, outfile, accession_ids, site_specificity)

# EasyPQP TargetedFileConverter
@cli.command()
@click.option('--in', 'infile', required=True, type=click.Path(exists=True), help='Transition list to convert.')
@click.option('--in_type', default=None, show_default=True, type=str, help='Input file type. Default: None, will be determined from input file. Valid formats: ["tsv", "mrm" ,"pqp", "TraML", "parquet"]')
@click.option('--out', 'outfile', required=False, default="library.pqp", show_default=True, type=click.Path(exists=False), help='Output file to be converted to.')
@click.option('--out_type', default=None, show_default=True, type=str, help='Output file type. Default: None, will be determined from output file. Valid formats: ["tsv", "pqp", "TraML"]')
@click.option('--legacy_traml_id/--no-legacy_traml_id', show_default=True, default=True, help='PQP to TraML: Should legacy TraML IDs be used?')
def targeted_file_converter(infile, in_type, outfile, out_type, legacy_traml_id):
    """
    Convert different spectral libraries / transition files for targeted proteomics and metabolomics analysis.

    Can convert multiple formats to and from TraML (standardized transition format). The following formats are supported:\b\n\n

        - @ref OpenMS::TraMLFile "TraML" \b\n
        - @ref OpenMS::TransitionTSVFile "OpenSWATH TSV transition lists" \b\n
        - @ref OpenMS::TransitionPQPFile "OpenSWATH PQP SQLite files" \b\n
        - SpectraST MRM transition lists \b\n
        - Skyline transition lists \b\n
        - Spectronaut transition lists \b\n
        - Parquet transition lists \b\n

    """
    converter = TargetedFileConverter(infile, outfile, in_type, out_type, legacy_traml_id)
    converter.convert()

# EasyPQP OpenSwathAssayGenerator
@cli.command()
@click.option('--in', 'infile', required=True, type=click.Path(exists=True), help="Input file (valid formats: 'tsv', 'mrm', 'pqp', 'TraML')")
@click.option('--in_type', default=None, show_default=True, type=str, help='Input file type. Default: None, will be determined from file extension or content. Valid formats: ["tsv", "mrm" ,"pqp", "TraML"]')
@click.option('--out', 'outfile', required=True, type=click.Path(exists=False), help="Output file (valid formats: 'tsv', 'pqp', 'TraML')")
@click.option('--out_type', default=None, show_default=True, type=str, help='Output file type. Default: None, will be determined from file extension or content. Valid formats: ["tsv", "mrm" ,"pqp", "TraML"]')
@click.option('--min_transitions', default=6, required=False, show_default=True, type=int, help='Minimal number of transitions')
@click.option('--max_transitions', default=6, required=False, show_default=True, type=int, help='Maximal number of transitions')
@click.option('--allowed_fragment_types', required=False, default='b,y', show_default=True, type=str, help='Allowed fragment types')
@click.option('--allowed_fragment_charges', required=False, default='1,2,3,4', show_default=True, type=str, help='Allowed fragment charge states')
@click.option('--enable_detection_specific_losses', required=False, type=bool, help='Set this flag if specific neutral losses for detection fragment ions should be allowed', default=False)
@click.option('--enable_detection_unspecific_losses', required=False, type=bool, help='Set this flag if unspecific neutral losses (H2O1, H3N1, C1H2N2, C1H2N1O1) for detection fragment ions should be allowed', default=False)
@click.option('--precursor_mz_threshold', default=0.025, show_default=True, required=False, type=float, help='MZ threshold in Thomson for precursor ion selection')
@click.option('--precursor_lower_mz_limit', default=400.0, show_default=True, required=False, type=float, help='Lower MZ limit for precursor ions')
@click.option('--precursor_upper_mz_limit', default=1200.0, show_default=True, required=False, type=float, help='Upper MZ limit for precursor ions')
@click.option('--product_mz_threshold', default=0.025, show_default=True, required=False, type=float, help='MZ threshold in Thomson for fragment ion annotation')
@click.option('--product_lower_mz_limit', default=350.0, show_default=True, required=False, type=float, help='Lower MZ limit for fragment ions')
@click.option('--product_upper_mz_limit', default=2000.0, show_default=True, required=False, type=float, help='Upper MZ limit for fragment ions')
@click.option('--swath_windows_file', 'swath_win', default=None, required=False, type=click.Path(exists=False), help="Tab separated file containing the SWATH windows for exclusion of fragment ions falling into the precursor isolation window: lower_offset upper_offset \newline400 425 \newline ... Note that the first line is a header and will be skipped. (valid formats: 'txt')")
@click.option('--unimod_file', 'unimod', required=False, type=click.Path(exists=False), help="(Modified) Unimod XML file (http://www.unimod.org/xml/unimod.xml) describing residue modifiability (valid formats: 'xml')")
@click.option('--enable_ipf', required=False, type=bool, help="IPF: set this flag if identification transitions should be generated for IPF. Note: Requires setting 'unimod_file", default=False)
@click.option('--max_num_alternative_localizations', default=10000, required=False, show_default=True, type=int, help='IPF: maximum number of site-localization permutations')
@click.option('--disable_identification_ms2_precursors', required=False, type=bool, help="IPF: set this flag if MS2-level precursor ions for identification should not be allowed for extraction of the precursor signal from the fragment ion data (MS2-level).", default=False)
@click.option('--disable_identification_specific_losses', required=False, type=bool, help="IPF: set this flag if specific neutral losses for identification fragment ions should not be allowed", default=False)
@click.option('--enable_identification_unspecific_losses', required=False, type=bool, help="IPF: set this flag if unspecific neutral losses (H2O1, H3N1, C1H2N2, C1H2N1O1) for identification fragment ions should be allowed", default=False)
@click.option('--enable_swath_specifity', required=False, type=bool, help="IPF: set this flag if identification transitions without precursor specificity (i.e. across whole precursor isolation window instead of precursor MZ) should be generated.", default=False)
def openswath_assay_generator(infile, in_type, outfile, out_type, min_transitions, max_transitions, allowed_fragment_types, allowed_fragment_charges, enable_detection_specific_losses, enable_detection_unspecific_losses, precursor_mz_threshold, precursor_lower_mz_limit, precursor_upper_mz_limit, product_mz_threshold, product_lower_mz_limit, product_upper_mz_limit, swath_win, unimod, enable_ipf, max_num_alternative_localizations, disable_identification_ms2_precursors, disable_identification_specific_losses, enable_identification_unspecific_losses, enable_swath_specifity):
    """
    Generates filtered and optimized assays for OpenSwathWorflow

    """
    assay_generator = OpenSwathAssayGenerator(infile, in_type, outfile, out_type, min_transitions, max_transitions, allowed_fragment_types, allowed_fragment_charges, enable_detection_specific_losses,
    enable_detection_unspecific_losses, precursor_mz_threshold, precursor_lower_mz_limit, precursor_upper_mz_limit, product_mz_threshold, product_lower_mz_limit, product_upper_mz_limit, swath_win, unimod, enable_ipf, max_num_alternative_localizations,
    disable_identification_ms2_precursors, disable_identification_specific_losses, enable_identification_unspecific_losses, enable_swath_specifity)

    
    assay_generator.read_input_file()
    assay_generator.annotate_transitions()
    assay_generator.write_output_file()
    

# EasyPQP OpenSwathDecoyGenerator
@cli.command()
@click.option('--in', 'infile', required=True, type=click.Path(exists=True), help='Input file to generate decoys for.')
@click.option('--in_type', default=None, show_default=True, type=str, help='Input file type. Default: None, will be determined from input file. Valid formats: ["tsv", "mrm" ,"pqp", "TraML"]')
@click.option('--out', 'outfile', required=False, default="library.pqp", show_default=True, type=click.Path(exists=False), help='Output file to be converted to.')
@click.option('--out_type', default=None, show_default=True, type=str, help='Output file type. Default: None, will be determined from output file. Valid formats: ["tsv", "pqp", "TraML"]')
@click.option('--method', default="shuffle", show_default=True, type=str, help='Decoy generation method. Valid methods: "shuffle", "pseudo-reverse", "reverse", "shift')
@click.option('--decoy_tag', default="DECOY_", show_default=True, type=str, help='Decoy tag.')
@click.option('--min_decoy_fraction', default=0.8, type=float, help='Minimum fraction of decoy / target peptides and proteins')
@click.option('--aim_decoy_fraction', default=1.0, type=float, help='Number of decoys the algorithm should generate (if unequal to 1, the algorithm will randomly select N peptides for decoy generation)')
@click.option('--shuffle_max_attempts', default=30, type=int, help='shuffle: maximum attempts to lower the amino acid sequence identity between target and decoy for the shuffle algorithm')
@click.option('--shuffle_sequence_identity_threshold', default=0.5, type=float, help='shuffle: target-decoy amino acid sequence identity threshold for the shuffle algorithm')
@click.option('--shift_precursor_mz_shift', default=0.0, type=float, help='shift: precursor ion MZ shift in Thomson for shift decoy method')
@click.option('--shift_product_mz_shift', default=20.0, type=float, help='shift: fragment ion MZ shift in Thomson for shift decoy method')
@click.option('--product_mz_threshold', default=0.025, type=float, help='MZ threshold in Thomson for fragment ion annotation')
@click.option('--allowed_fragment_types', default='b,y', type=str, help='allowed fragment types')
@click.option('--allowed_fragment_charges', default='1,2,3,4', type=str, help='allowed fragment charge states')
@click.option('--switchKR/--no_switchKR', 'switchkr', default=True, help='Whether to switch terminal K and R (to achieve different precursor mass)')
@click.option('--enable_detection_specific_losses/--disable_detection_specific_losses', default=False, help='set this flag if specific neutral losses for detection fragment ions should be allowed')
@click.option('--enable_detection_unspecific_losses/--disable_detection_unspecific_losses', default=False, help='set this flag if unspecific neutral losses (H2O1, H3N1, C1H2N2, C1H2N1O1) for detection fragment ions should be allowed')
@click.option('--separate/--no_separate', default=False, help='set this flag if decoys should not be appended to targets.')
def openswath_decoy_generator(infile, in_type, outfile, out_type, method, decoy_tag, min_decoy_fraction, aim_decoy_fraction, shuffle_max_attempts, shuffle_sequence_identity_threshold, shift_precursor_mz_shift, shift_product_mz_shift, product_mz_threshold, allowed_fragment_types, allowed_fragment_charges, switchkr, enable_detection_specific_losses, enable_detection_unspecific_losses, separate):
    """
    Generate decoys for spectral libraries / transition files for targeted proteomics and metabolomics analysis.

    Can generate decoys for multiple formats. The following formats are supported:\b\n\n

        - @ref OpenMS::TraMLFile "TraML" \b\n
        - @ref OpenMS::TransitionTSVFile "OpenSWATH TSV transition lists" \b\n
        - @ref OpenMS::TransitionPQPFile "OpenSWATH PQP SQLite files" \b\n

    """
    generator = OpenSwathDecoyGenerator(infile, outfile, in_type, out_type, method, decoy_tag, min_decoy_fraction, aim_decoy_fraction, shuffle_max_attempts, shuffle_sequence_identity_threshold, shift_precursor_mz_shift, shift_product_mz_shift, product_mz_threshold, allowed_fragment_types, allowed_fragment_charges, switchkr, enable_detection_specific_losses, enable_detection_unspecific_losses, separate)
    generator.generate_decoys()
